import torch
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models._numpy.NumpySurvivalModel import NumpySurvivalModel

class Weibull(TorchSurvivalModel):
    def __init__(self) -> None:
        super().__init__()

    def _sample(self, forward_out, num_samples, generator=None):
        return torch.distributions.Weibull(forward_out[:,1], forward_out[:,0]).sample(num_samples)
    
    def max_time(self):
        return torch.inf
    
    def _inv_survival_function(self, forward_out, S):
        return forward_out[:, 1]*(-torch.log(S))**(1/forward_out[:, 0])

    def inv_survival_function(self, X, S):
        return self._inv_survival_function(self.forward(X), S)

    def _survival_function(self, forward_out, t):
        return torch.exp(-((t / forward_out[:,1]) ** forward_out[:,0]))

    def _density_function(self, forward_out, t):
        return (
            (forward_out[:,0] / forward_out[:,1])
            * (t / forward_out[:,1]) ** (forward_out[:,0] - 1)
            * torch.exp(-((t / forward_out[:,1]) ** forward_out[:,0]))
        )
    
    def _sample(self, forward_out, num_samples=1, generator=None):
        u = torch.rand((forward_out.shape[0]*num_samples,),
                       device=forward_out.device,
                       dtype=forward_out.dtype,
                       generator=generator)
        return self._inv_survival_function(
            forward_out.repeat_interleave(num_samples, 0),
            u).reshape(-1, num_samples)

    def forward(self, X):
        return X



import torch
import pysare
import numpy as np
import matplotlib.pyplot as plt
import torch.utils.data
import pysare
import numpy as np
import matplotlib.pyplot as plt
import lifelines as lf
import pandas as pd
from collections.abc import Iterable
from matplotlib.axes._axes import Axes
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models._numpy.NumpySurvivalModel import NumpySurvivalModel
from pysare.models._numpy.NumpySurvivalModel import FrozenNumpySurvivalModel
from pysare.data import Dataset
from torch.utils.data.dataloader import DataLoader
from typing import Union, Optional, Tuple, Literal
from numpy.typing import ArrayLike

from pysare._typing import X, T, E


def binomial_log_likelihood(model: Union[TorchSurvivalModel, NumpySurvivalModel],
                            data: Union[Tuple[X, T, E], Dataset, DataLoader],
                            times: Union[T, int],
                            integrated: bool = False,
                            censoring: Union[Literal['KM', 'none'],
                                             FrozenNumpySurvivalModel] = 'KM',
                            plot: Union[bool, Axes] = False) -> pd.DataFrame:
    r"""Binomial log-likelihood

    Calculates the binomial log-likelihood score and its integrated counterpart.
    Also considers censoring, see [1] for more information.

    Parameters
    ----------
        model : SurvivalModel or NumpySurvivalModel
            Model to be evaluated. If data is a tuple (X, T, E) of non-tensors, 
            the model should be a NumpySurvivalModel, and otherwise it should 
            be a SurvivalModel.
        data : tuple(X,T,E) or Dataset or Dataloader
            The data to be used for evaluation. If data is a tuple (X, T, E) of 
            non-tensors, the model should be a NumpySurvivalModel, and otherwise 
            it should be a SurvivalModel.
        times : int or (n,) array_like
            Times where the score is evaluated. Should have same data type as 
            the times (T) in the data parameter. If int, this number of linearly
            spaced times between zero and largest time in dataset is used. Times
            where the survival function of the censoring distribution is zero 
            are removed to avoid NaN in the output.
        integrated : bool, default=False
            Indicates if the integrated score should be calculated, False by 
            default.
        censoring : {'KM' , 'none'} or FrozenNumpySurvivalModel, default='KM'
            Determines how the censoring distribution is calculated. If 'KM' 
            it is estimated using Kaplan-Meier estimator using the data, if 
            'None' no censoring is use, and if it is a FrozenNumpySurvivalModel
            this model is used.
        plot : bool or Axis, default=False
            If True, a plot of the result will be created after the calculation,
            and if an Axis is passed the plot wil be done in this axis. 
            If False (default), no plot is made.

    Returns
    ----------
        result : DataFrame
            A pandas DataFrame with the result.

    References
    ----------
    [1] Graf, Erika, et al. "Assessment and comparison of prognostic 
    classification schemes for survival data." Statistics in medicine 
    18.17‐18 (1999): 2529-2545.

    """

    if hasattr(model, 'eval'):
        model.eval()

    # Interpret inputs
    # ==========================================================================
    if isinstance(data, tuple):  # data is a tuple (X, T, E)
        dataloader = (data,)
    elif isinstance(data, Iterable):  # data is a DataLoader
        dataloader = data
    else:  # Data is a Dataset
        dataloader = (data[:],)

    dataset = data

    # Full vector of times  and events
    # ===========================================================================
    T = []
    E = []

    for X_batch, T_batch, E_batch in dataloader:
        T.append(T_batch.reshape(-1,))
        E.append(E_batch.reshape(-1,))

    if torch.is_tensor(T[0]):
        T = torch.concat(T)
        E = torch.concat(E)
        T_np = np.array(T)
        E_np = np.array(E)
    else:
        T = np.concatenate(T)
        E = np.concatenate(E)
        T_np = T
        E_np = E

    if isinstance(times, int):
        if torch.is_tensor(T):
            times = torch.linspace(
                0, max(T), times, dtype=T.dtype, device=T.device)
            times_np = np.array(times.cpu())
        else:
            times = np.linspace(0, max(T), times)
            times_np = times
    else:
        if torch.is_tensor(times):
            times_np = np.array(times.cpu())
        else:
            times_np = np.array(times)
            times = torch.tensor(times, dtype=T.dtype, device=T.device)

    if isinstance(model, pysare.models.TorchSurvivalModel):
        def get_time_vector(time, length):
            return time.repeat(length)
    else:
        def get_time_vector(time, length):
            return time.repeat(length).reshape(-1, 1)

    if censoring == "KM":
        KMC = lf.KaplanMeierFitter()
        KMC.fit(T_np, ~E_np)
        S_cens_t = KMC.survival_function_at_times(times_np).values
        S_cens_T_inv = 1 / \
            KMC.survival_function_at_times(
                T_np.reshape(-1,)).values.reshape(-1,)
    elif censoring is False:
        S_cens_t = np.ones_like(times_np)
        S_cens_T_inv = np.ones((T.shape[0],))
    else:
        S_cens_t = censoring.survival_function(times)
        S_cens_T_inv = 1/censoring.survival_function(T)

    inds = S_cens_t > 0.0
    times = times[inds]
    times_np = times_np[inds]
    S_cens_t = S_cens_t[inds]


    B = np.zeros_like(times_np)

    # Function such that ind*log = 0 if ind==false and log==-inf
    def indicator_times_log(ind, log):
        log[~ind] = 0
        return log

    with torch.no_grad():
        for n in range(times_np.shape[0]):

            # R = model.survival_probability(X, t[n]).reshape(-1,)
            R = []
            for X_batch, _, _ in dataloader:
                R.append(np.array(model.survival_function(
                    X_batch, get_time_vector(times[n:n+1],X_batch.shape[0])).reshape(-1,), copy=False))
            R = np.concatenate(R)

            B1 = (indicator_times_log(
                (T_np <= times_np[n]), np.log(1-R))*S_cens_T_inv)[E]

            B[n] = (B1.sum()
                    + indicator_times_log((T_np > times_np[n]), np.log(R)).sum()/S_cens_t[n])/T_np.shape[0]

    if integrated:
        IB = np.empty_like(B)
        IB[0] = 0.0
        IB[1:] = (B[1:]+B[:-1])*np.diff(times_np)/2
        IB[1:] = IB.cumsum()[1:]/(times_np[1:]-times_np.min())

    # Plot if ax is not False
    if plot:
        color_1, color_2 = plt.rcParams['axes.prop_cycle'].by_key()[
            'color'][:2]

        if plot is True:
            if censoring is not False:
                fig, ax = plt.subplots(nrows=2)
            else:
                fig, ax = plt.subplots(nrows=1)
        else:
            ax = plot

        if censoring is not False:
            if censoring == 'KM':
                KMC.plot_survival_function(ax=ax[1])
            else:
                ax[1].plot(t, S_cens_t)
            ax[1].legend(['Kaplan-Meier Estimate',
                         'Confidence Intervall (95%)'])
            ax[1].set_xlabel('Time')
            ax[1].set_ylabel('Censoring Distribution')
        else:
            ax = [ax]

        ax[0].plot(times_np, B, color=color_1)
        if integrated:
            ax0_twin = ax[0].twinx()
            ax0_twin.plot(times_np, IB, color=color_2)
            ax0_twin.set_ylabel('Int. Binomial log-lik.')

            ax[0].yaxis.label.set_color(color_1)
            ax0_twin.yaxis.label.set_color(color_2)
            ax[0].tick_params(axis='y', colors=color_1)
            ax0_twin.tick_params(axis='y', colors=color_2)

        ax[0].set_ylabel('Binomial log-lik.')
        ax[0].set_xlabel('Time')
        plt.tight_layout()

    if integrated:
        return pd.DataFrame(data={'binomial_log_lik': B, 'int_binomial_log_lik': IB,
                                  'time': times_np, 'censoring_survival': S_cens_t}).set_index('time')
    else:
        return pd.DataFrame(data={'binomial_log_lik': B, 'time': times_np, 'censoring_survival': S_cens_t}).set_index('time')



import torch
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models.piecewise.grid import DiscretizationGrid
from pysare.models.utils.MLP import MLP
from abc import abstractmethod


class ConstantHazard(TorchSurvivalModel):
    r"""Implements an piecewise constant hazard model.
    
    The forward output size should be 
        number of intervals + potential extra inputs required by the grid

    See :mod:`pysare.models.piecewise` for information on how to use 
    piecewise models.
    """
    def __init__(self, grid: DiscretizationGrid):
        super(ConstantHazard, self).__init__()
        self.grid = grid

    def _log_likelihood(self, z, T, E):
        disc, z = self.grid.bin(T, z)
        N = T.shape[0]

        e = torch.exp(z)

        ln_e = z

        dH = e*disc.dt
        H_n = dH.cumsum(dim=1)
        h = e[range(N), disc.ind]
        H = H_n[range(N), disc.ind] - (disc.tau_next_ind-T)*h
        # S = torch.exp(-H)
        # f = h*S

        ind_f = torch.where(E)[0]
        ind_c = torch.where(~E)[0]
        ln_f = ln_e[range(N), disc.ind] - H
        ln_S = -H
        log_likelihood = ln_f
        log_likelihood[ind_c] = ln_S[ind_c]

        return log_likelihood

    def max_time(self):
        return self.grid.max_time()
    
    def forward_output_size(self):
        return self.grid.num_intervals()+self.grid.extra_forward_output_size()

    def _density_function(self, z, T):
        N = T.shape[0]
        disc, z = self.grid.bin(T, z)

        e = torch.exp(z)

        

        dH = e*disc.dt
        H_n = dH.cumsum(dim=1)
        h = e[range(N), disc.ind]
        H = H_n[range(N), disc.ind] - (disc.tau_next_ind-T)*h
        S = torch.exp(-H)
        # f = h*S
        return h*S

    def _survival_function(self, z, T):
        N = T.shape[0]
        disc, z = self.grid.bin(T, z)
        e = torch.exp(z)

        

        dH = e*disc.dt
        H_n = dH.cumsum(dim=1)
        h = e[range(N), disc.ind]
        H = H_n[range(N), disc.ind] - (disc.tau_next_ind-T)*h
        S = torch.exp(-H)
        return S

    def forward(self, X):
        return X

    @classmethod
    def MLP_implementation(cls, grid, input_size, hidden_sizes,
                           activation=torch.nn.ReLU, dropout=False,
                           batch_norm=False):
        return MLPConstantHazard(grid, input_size, hidden_sizes,
                                 activation, dropout,
                                 batch_norm)


class MLPConstantHazard(ConstantHazard):
    def __init__(self, grid, input_size, hidden_sizes,
                 activation=torch.nn.ReLU, dropout=0.0, batch_norm=False):
        super(MLPConstantHazard, self).__init__(grid=grid)

        self.input_size = input_size

        self.layers = MLP(input_size, grid.num_intervals() + grid.extra_forward_output_size(),
                          hidden_sizes, activation, dropout, batch_norm)

    def forward(self, X):
        return self.layers(X)



import numpy as np
import torch

class KaplanMeierEstimator():
    def __init__(self, T, E=None) -> None:

        T = np.array(T)
        order = T.argsort()
        T = T[order]

        if E is None:
            E = np.ones_like(T, dtype=bool)
        else:
            E = np.array(E, dtype=bool)[order]

        first_unique_ind = np.where(
            np.diff(T, prepend=T[0]-1.0, append=T[-1]+1.0) > 0)[0]
        num_prev_failures = np.zeros(len(T)+1, dtype=int)
        np.cumsum(E, out=num_prev_failures[1:])

        num_failures = np.diff(num_prev_failures[first_unique_ind])
        num_at_risc = len(T)-first_unique_ind[:-1]

        have_failure = num_failures > 0
        step_inds = first_unique_ind[:-1][have_failure]

        num_failures = num_failures[have_failure]
        num_at_risc = num_at_risc[have_failure]


        if T[step_inds[-1]]==T[-1]:
            self.times = T[step_inds]
            Q = (1-num_failures/num_at_risc)
            self.survival_probabilities = np.cumprod(Q)
        else:
            self.times = np.empty(len(step_inds)+1)
            self.times[:-1] = T[step_inds]
            self.times[-1] = T[-1]
            self.survival_probabilities = np.empty_like(self.times)
            self.survival_probabilities[-1] = 0.0
            Q = (1-num_failures/num_at_risc)
            self.survival_probabilities[:-1] = np.cumprod(Q) 

        self.probabilities = np.diff(self.survival_probabilities, prepend=1.0)

    def survival_function(self, T):
        T = np.array(T)
        inds = np.atleast_1d(np.searchsorted(self.times,T,side="right")-1)
        
        probs = self.survival_probabilities[inds]
        probs[inds<0] = 1.0
        return probs.reshape(T.shape)
    
    def sample(self, size=None, generator=None):
        
        u = torch.rand(size=tuple(np.atleast_1d(size)),
                   dtype=torch.float64,
                   generator=generator).numpy()

        inds = (self.survival_probabilities.shape[0] - 
                np.searchsorted(np.flip(self.survival_probabilities), u))
        
        return self.times[inds].reshape(size)


import numpy as np
import matplotlib.pyplot as plt


def midpoints(t, y):
    step_inds = np.where(np.diff(y[1:-1]) != 0)[0]+1
    y = np.insert(0.5*(y[step_inds]+y[step_inds+1]),
                  [0, len(step_inds)], y[[0, -1]])
    t = np.insert(0.5*(t[step_inds]+t[step_inds+1]),
                  [0, len(step_inds)], t[[0, -1]])
    return t, y


def interpolate(t, y, samples_per_tu):

    num_samples = max(np.ceil(samples_per_tu*(t[-1]-t[0])).astype(int), 2)

    if num_samples < len(t):
        t_plot = np.linspace(*t[[0, -1]], num_samples)
        y_plot = np.interp(t_plot, t, y)

        return t_plot, y_plot
    else:
        return t, y


def get_segments(t, y, resolution, tol):
    # Position of midpoint in relation to the adjacent
    x = (t[1:-1]-t[:-2])/(t[2:]-t[:-2])

    isinf = np.isinf(y)
    becomes_inf = np.diff(isinf) != 0

    y_hat = y[2:]*x + (1.0-x)*y[:-2]

    e = (np.abs(y_hat-y[1:-1])-resolution).clip(0.0)/y[1:-1]
    e[np.isnan(e)] = 2*tol
    is_discontinuous = (np.diff((e > tol).astype(
        int), prepend=e[0]) == 1) | becomes_inf[1:]

    disc_inds = np.where(is_discontinuous)[0]+2

    segments = np.insert(disc_inds, [0, len(disc_inds)], [0, len(t)])
    return segments


def find_nearest(data, keys):
    inds = np.atleast_1d(np.searchsorted(
        data, keys, side='left').clip(1, data.shape[0]-1))

    inds[keys-data[inds-1] < data[inds]-keys] -= 1

    return inds.reshape(keys.shape)


class HorizonBasedPolicy():
    def __init__(self,
                 horizon,
                 replacement_age,
                 expected_cost,
                 survival_probability,
                 preventive_cost=None,
                 failure_cost=None,
                 grid_spacing=0.0,
                 interpolation='nearest',
                 ) -> None:
        self.horizon = horizon
        self._replacement_age = replacement_age
        self.expected_cost = expected_cost
        self.survival_probability = survival_probability
        self.preventive_cost = preventive_cost
        self.failure_cost = failure_cost
        self.grid_spacing = grid_spacing
        self.interpolation = interpolation

    def __getitem__(self, horizons):

        horizons = np.array(horizons, dtype=float)
        horizons[horizons < 0] = np.nan
        horizons[horizons > self.horizon[-1]+0.5*self.grid_spacing] = np.nan

        if self.interpolation == 'nearest':
            inds = find_nearest(self.horizon, horizons)
            replacement_ages = self._replacement_age[inds]
        if self.interpolation == 'linear':
            replacement_ages = np.interp(
                horizons, self.horizon, self._replacement_age)

        replacement_ages = np.atleast_1d(replacement_ages)
        replacement_ages[replacement_ages == 0] = self.horizon[-1]
        return replacement_ages.reshape(horizons.shape)

    def evaluate(self,
                 failure_distribution,
                 preventive_cost=None,
                 failure_cost=None):
        if preventive_cost is None:
            preventive_cost = self.preventive_cost

        if failure_cost is None:
            failure_cost = self.failure_cost

        return _evaluate_sequential_finite_horizon(
            failure_distribution.survival_function(self.horizon),
            self.horizon,
            self._replacement_age,
            preventive_cost,
            failure_cost)

    def simulate(self,
                 failure_distribution,
                 num_simulations=1,
                 preventive_cost=None,
                 failure_cost=None,
                 horizon=None
                 ):
        if preventive_cost is None:
            if self.preventive_cost is None:
                raise ValueError(
                    'Preventive cost not set for policy and must be provided')
            preventive_cost = self.preventive_cost

        if failure_cost is None:
            if self.failure_cost is None:
                raise ValueError(
                    'Failure cost not set for policy and must be provided')
            failure_cost = self.failure_cost

        if horizon is None:
            horizon = self.horizon[-1]

        remaining_horizon = np.full((num_simulations,), horizon)
        num_preventive = np.zeros_like(remaining_horizon, dtype=int)
        num_failures = np.zeros_like(remaining_horizon, dtype=int)

        active = np.arange(0, num_simulations)

        while len(active) > 0:
            failure_time = failure_distribution.sample(
                len(active)).reshape(active.shape)

            replacement_age = self[remaining_horizon[active]]

            is_preventive = failure_time > replacement_age
            is_failure = np.logical_not(is_preventive)

            replacement_time = failure_time
            replacement_time[is_preventive] = replacement_age[is_preventive]

            remaining_horizon[active] -= replacement_time

            finished = remaining_horizon[active] <= 0

            num_preventive[active[~finished & is_preventive]] += 1
            num_failures[active[~finished & is_failure]] += 1

            active = active[~finished]

        num_preventive = num_preventive.squeeze()
        num_failures = num_failures.squeeze()

        cost = preventive_cost*num_preventive + failure_cost*num_failures
        return {'cost': cost,
                'num_preventive': num_preventive,
                'num_failures': num_failures}

    def plot_policy(self,
                    ax=None,
                    color=None,
                    linewidth=None,
                    linestyle=None,
                    draw_inf=True,
                    resolution=None,
                    use_midpoints=True,
                    discontinuity_tolerance=1e-6):
        if ax is None:
            _, ax = plt.subplots(1, 1)

        if resolution is not None:
            samples_per_tu = resolution/(self.horizon[-1]-self.horizon[0])
        else:
            samples_per_tu = np.inf

        h = self.horizon
        r = self._replacement_age.copy()
        is_inf = (r == 0.0) | (r >= h)

        if not is_inf.all():
            r_max = np.nanmax(r[~is_inf])
        else:
            r_max = 0.0
            
        # Something big to make sure it becomes a segment
        r[is_inf] = np.inf
        segments = get_segments(
            h, r, self.grid_spacing, discontinuity_tolerance)

        if isinstance(draw_inf, bool):
            r[is_inf] = r_max*1
        else:
            r[is_inf] = float(draw_inf)

        def plot_segment(n, color):

            if is_inf[segments[n]]:

                if draw_inf is not False:
                    inds = [segments[n], segments[n+1]-1]
                    line, = ax.plot(h[inds], r[inds], '--',
                                    color=color, linewidth=linewidth)
                else:
                    return color
            else:
                inds = slice(segments[n], segments[n+1])
                if use_midpoints:
                    hn, rn = midpoints(h[inds], r[inds])
                else:
                    hn, rn = h[inds], r[inds]

                if samples_per_tu < np.inf:
                    hn, rn = interpolate(hn, rn, samples_per_tu)

                line, = ax.plot(hn, rn, color=color, linewidth=linewidth, linestyle=linestyle)

            if color is None:
                return line.get_color()
            else: 
                return color

        color = plot_segment(0, color)

        for n in range(1, len(segments)-1):
            color = plot_segment(n, color)
        return ax, segments


def _evaluate_sequential_finite_horizon(S, h, r, c_p, c_f):
    C = np.zeros_like(S)
    # r = np.zeros_like(S)
    N = len(C)
    dF = -np.diff(S)
    dh = h[1]-h[0]

    for n in range(1, N):
        if (r[n] <= 0) or (r[n] >= n*dh):
            JC = np.flip(C[:n])
            V = np.sum((c_f + JC)*dF[:n])
            C[n] = V
        else:
            k = max(1, int(np.round(r[n]/dh))-1)

            JC = np.flip(C[n-k-1:n])
            V = np.sum((c_f + JC)*dF[:k+1])
            W = (c_p+JC[-1])*S[k] + V
            C[n] = W

    return HorizonBasedPolicy(h, r, C, S, grid_spacing=dh, preventive_cost=c_p, failure_cost=c_f)


def _sequential_finite_horizon(S, h, c_p, c_f):
    C = np.zeros_like(S)
    r = np.zeros_like(S)
    N = len(C)
    dF = -np.diff(S)
    dh = h[1]-h[0]

    for n in range(1, N):
        JC = np.flip(C[:n])
        V = np.cumsum((c_f + JC)*dF[:n])
        W = (c_p+JC)*S[:n] + V

        k_opt = W.argmin()
        W_opt = W[k_opt]

        if W_opt < V[-1]:


            ind = np.where(W-W_opt <= np.spacing(W_opt))[0][-1]
            r[n] = (ind+1)*dh
            C[n] = W_opt
        else:
            r[n] = 0.0
            C[n] = V[-1]

    return HorizonBasedPolicy(h, r, C, S, grid_spacing=dh, preventive_cost=c_p, failure_cost=c_f)


def sequential_finite_horizon(failure_distribution,
                              preventive_cost,
                              failure_cost,
                              horizon,
                              num_disc_points):
    if hasattr(failure_distribution, 'to_numpy'):
        failure_distribution = failure_distribution.to_numpy()

    h = np.linspace(0.0, horizon, num_disc_points)

    # S = np.zeros_like(h)
    # S[h <= failure_distribution.max_time()] = failure_distribution.survival_function(
    #     h[h <= failure_distribution.max_time()])
    S = failure_distribution.survival_function(h)

    return _sequential_finite_horizon(S, h, preventive_cost, failure_cost)



import torch
import numpy as np
from abc import ABC, abstractmethod


class EnergyIntegrator:
    r"""Base class for energy integrators."""

    def log_integrate(self, model, X, T, t_m, tail_ratio=1.2):
        pass

    @abstractmethod
    def integrate(self, model, X, T, t_m, tail_ratio=1.2):
        pass


class RandomMonteCarlo(EnergyIntegrator):
    r"""Integrator based on Monte-Carlo integration.

    Implements Monte-Carlo integration based on uniform sampling of the 
    integration interval.

    Parameters
    ----------
    num_samples : int
        Number of random samples in the integration
    """
    def __init__(self, num_samples: int):
        super(RandomMonteCarlo, self).__init__()
        self.N = num_samples

    def log_integrate(self, model, X, T, t_m, tail_ratio=1.2):

        return torch.log(self.integrate(model, X, T, t_m, tail_ratio=1.2))

    def integrate(self, model, X, T, t_m, tail_ratio=1.2):

        X_tail = torch.cat((X, T.reshape(-1, 1)*0+tail_ratio*t_m), dim=1)
        F_tail = torch.exp(model.energy(X_tail)).reshape(-1, 1)

        T = T.reshape(-1, 1)
        if X.dim() == 1:
            X = X.reshape(-1, 1)

        dt = torch.kron(t_m-T, torch.ones((self.N, 1)))
        t = torch.rand((self.N*X.shape[0], 1))

        X = torch.cat((torch.kron(X, torch.ones((self.N, 1))),
                       dt.reshape(-1, 1)*t.reshape(-1, 1)
                       + torch.kron(T.reshape(-1, 1), torch.ones((self.N, 1)))
                       ), dim=1)

        F = torch.exp(model.energy(X)).reshape(-1, self.N)

        Z = F.sum(axis=1).reshape(-1, 1)/self.N*(t_m-T) \
            + F_tail*(tail_ratio-1)*t_m

        return Z


class EquidistantTrapezoidal(EnergyIntegrator):
    r"""Integrator based on the Trapezoidal rule.

    Implements the Trapezoidal rule on an equidistant grid.

    Parameters
    ----------

    num_points : int
        Number of points in the integration.
    """
    def __init__(self, num_points: int):
        super(EquidistantTrapezoidal, self).__init__()
        self.N = num_points

    def _segment(self, X, T, t_m, tail_ratio):
        if X.dim() == 1:
            X = X.reshape(-1, 1)

        u = torch.ones((self.N,))
        u = u / u.sum()
        #        t = torch.cat((torch.zeros(1), u.cumsum(dim=0)))

        t = torch.cat((torch.zeros(1), u.cumsum(
            dim=0), torch.ones(1) * tail_ratio))
        # u = torch.cat((u, torch.ones(1) * (self.tail_ratio - 1.)))
        # FIXME: tail not handled correctly below
        X = torch.cat((X.repeat_interleave(self.N+2,0),
                       torch.kron(t_m - T.reshape(-1, 1), t.reshape(-1, 1))
                       + torch.kron(T.reshape(- 1, 1),
                                    torch.ones_like(t.reshape(-1, 1)))
                       ), dim=1)
        X[self.N + 1::self.N + 2, -1] = tail_ratio*t_m

        return X, u

    def integrate(self, model, X, T, t_m, tail_ratio=1.2):
        X, u = self._segment(X, T, t_m, tail_ratio)

        F = torch.exp(model.energy(X)).reshape(-1, self.N + 2)

        Z = ((F[:, :-2] @ u + F[:, 1:-1] @ u) / 2).reshape(-1, 1) \
            * (t_m - T.reshape(- 1, 1)).reshape(-1, 1) \
            + F[:, -1].reshape(-1, 1) * t_m * (tail_ratio - 1)

        return Z

    def log_integrate(self, model, X, T, t_m, tail_ratio=1.2):
        X, u = self._segment(X, T, t_m, tail_ratio)

        f = model.energy(X).reshape(-1, self.N + 2)
        f_ast = f.max(dim=1).values
        f -= f_ast[:, None]
        F = torch.exp(f)

        log_Z = f_ast.reshape(-1, 1) \
            + torch.log(((F[:, :-2] @ u
                          + F[:, 1:-1] @ u) / 2).reshape(- 1, 1)
                        * (t_m - T.reshape(-1, 1)).reshape(- 1, 1)
                        + F[:, -1].reshape(-1, 1) * t_m * (tail_ratio - 1))

        return log_Z


# Simpson coefficients
_W_I = (1/6, 4/6, 1/6)
# Simpson-Trapetsoidal error coefficients
_W_e = (1/12, -1/6, 1/12)

class AdaptiveTrapezoidalSimpsons(EnergyIntegrator):
    r"""Adaptive integrator combining Trapezoidal and Simpson's rule.

    Implements an adaptive integration scheme where the difference between
    the Trapezoidal and Simpson's rule on each segment is used to estimate
    the error on each segment. The method iteratively bisects each interval
    that does not satisfy at least one of the tolerances, starting with the 
    initial integration interval.

    Parameters
    ----------
    
    rtol : float>0, default=1e-4
        Relative tolerance (satisfied if error < value*rtol).
    atol : float>0, default=1e-5
        Absolute tolerance (satisfied if error < rtol).
    """
    def __init__(self, rtol: float = 1e-4, atol: float = 1e-5):
        super(AdaptiveTrapezoidalSimpsons, self).__init__()

        self.rtol = rtol
        self.atol = atol

        # Simpson coefficients
        self.W_I = torch.tensor(_W_I)
        # Simpson-Trapetsoidal error coefficients
        self.W_e = torch.tensor(_W_e)

    def integrate(self, model, X, a, b, tail_ratio=1.2):
        a = a.reshape(-1, 1)
        b = torch.ones_like(a)*b

        rtol_pu = self.rtol/(b-a)
        atol_pu = self.atol/(b-a)

        if X.dim() == 1:
            X = X.reshape(-1, 1)

        X_ = torch.cat((X, b*tail_ratio), dim=1)

        # Contribution from tail
        Z = torch.exp(model.energy(X_)) * b * (tail_ratio - 1)

        def f(t, inds=slice(None)):
            X_[inds, -1] = t.view(-1,)
            return torch.exp(model.energy(X_[inds]))

        fl = f(a)
        fh = f(b)
        N = len(fl)

        not_evaluated = [((a, b), (fl, fh), torch.arange(0, N))]

        while not_evaluated:
            intervall = not_evaluated.pop()
            a, b = intervall[0]

            fl, fh = intervall[1]
            inds = intervall[2]

            while True:
                m = (a+b)/2
                fm = f(m, inds)
                dx = (b-a)

                f_stack = torch.concat((fl, fm, fh), dim=1)

                I_intervall = dx*(f_stack@self.W_I).reshape(-1, 1)
                e = torch.abs(dx**2*(f_stack@self.W_e).reshape(-1, 1))
                accurate = ((e < atol_pu[inds]) | (
                    e < I_intervall[0]*rtol_pu[inds])).reshape(-1,)

                if not torch.all(accurate):

                    Z[inds[accurate]] += I_intervall[accurate]
                    not_evaluated.append(((m[~accurate], b[~accurate]),
                                          (fm[~accurate], fh[~accurate]),
                                          inds[~accurate]))

                    inds = inds[~accurate]
                    b = m[~accurate]
                    a = a[~accurate]
                    fl = fl[~accurate]
                    fh = fm[~accurate]
                else:
                    Z[inds] += I_intervall
                    break

        return Z

    def log_integrate(self, model, X, T, t_m, tail_ratio=1.2):
        return torch.log(self.integrate(model, X, T, t_m, tail_ratio))



from typing import Iterable
from math import floor
import pysare
import numpy as np
import torch
import matplotlib.pyplot as plt
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models._numpy.NumpySurvivalModel import NumpySurvivalModel, FrozenNumpySurvivalModel

def ensemble(models):
    for model in models:
        if isinstance(model, FrozenNumpySurvivalModel):
            return FrozenNumpyEnsembleModel(models)
        elif isinstance(model, NumpySurvivalModel):
            return NumpyEnsembleModel(models)
        else:
            return TorchEnsembleModel(models)


class TorchEnsembleModel(TorchSurvivalModel):
    def __init__(self, models):
        super().__init__()

        self.models = torch.nn.ModuleList()
        for model in models:
            if isinstance(model, TorchSurvivalModel):
                self.models.append(model)
            else:
                raise TypeError('One of the entries in `models`'
                                + ' is no a TorchSurvivalModel')

    def _all_survival_functions(self, X, T):
        return torch.stack([model.survival_function(X, T) for
                            model in self.models])

    def _all_density_functions(self, X, T):
        return torch.stack([model.density_function(X, T) for
                            model in self.models])

    def survival_function(self, X: torch.Tensor, T: torch.Tensor) -> torch.Tensor:
        return self._all_survival_functions(X, T).mean(dim=0)

    def density_function(self, X: torch.Tensor, T: torch.Tensor) -> torch.Tensor:
        return self._all_density_functions(X, T).mean(dim=0)

    def survival_confidence(self, X, T, confidence_level=0.95, return_survival=False):

        return self._calculate_confidence(
            self._all_survival_functions(X, T).sort(dim=0)[0],
            confidence_level,
            return_survival)

    def density_confidence(self, X, T, confidence_level=0.95, return_survival=False):
        return self._calculate_confidence(
            self._all_density_functions(X, T).sort(dim=0)[0],
            confidence_level,
            return_survival)

    def _calculate_confidence(self, quantity, confidence_level, return_mean):

        alpha_0 = .5*(1.0 - (len(self.models)-2)/(max(1, len(self.models)-1)))

        if isinstance(confidence_level, Iterable):
            alpha = .5 * \
                (1.0-torch.tensor(confidence_level, dtype=torch.float64))
            alpha_index = (alpha/alpha_0).to(dtype=torch.int64)
        else:
            alpha = .5*(1.0-confidence_level)
            alpha_index = int(alpha/alpha_0)

        if return_mean:
            return quantity[alpha_index, :], quantity[-1-alpha_index, :], quantity.mean(dim=0)
        else:
            return quantity[alpha_index, :], quantity[-1-alpha_index, :]

    def forward(self, X):
        return torch.stack([model(X) for model in self.models])


class NumpyEnsembleModel():
    def __init__(self, models) -> None:
        super().__init__()

        self.models = []
        for model in models:
            if isinstance(model, NumpySurvivalModel):
                self.models.append(model)
            else:
                raise TypeError('One of the entries in `models`'
                                + ' is no a NumpySurvivalModel')

    def _all_survival_functions(self, X, T):
        return np.stack(
            [model.survival_function(X,T).reshape(len(X),-1)
             for model in self.models])
    def _all_density_functions(self, X, T):
        return np.stack(
            [model.density_function(X,T).reshape(len(X),-1)
             for model in self.models])

    def survival_function(self,X,T):
        return self._all_survival_functions(X,T).mean(axis=0).squeeze()

    def density_function(self,X,T):
        return self._all_density_functions(X,T).mean(axis=0).squeeze()

    def survival_confidence(self, X, T, confidence_level=0.95, 
                            return_survival=False):
        S = self._all_survival_functions(X, T)
        S.sort(axis=0)
        return self._calculate_confidence(
            S,
            confidence_level,
            return_survival)

    def density_confidence(self, X, T, confidence_level=0.95, 
                            return_density=False):
        S = self._all_density_functions(X, T)
        S.sort(axis=0)
        return self._calculate_confidence(
            S,
            confidence_level,
            return_density)


    def _calculate_confidence(self, quantity, confidence_level, return_mean):

        alpha_0 = .5*(1.0 - (len(self.models)-2)/(max(1, len(self.models)-1)))

        if isinstance(confidence_level, Iterable):
            alpha = .5 * \
                (1.0-np.array(confidence_level))
            alpha_index = (alpha/alpha_0).astype(int)
        else:
            alpha = .5*(1.0-confidence_level)
            alpha_index = int(alpha/alpha_0)

        if return_mean:
            return quantity[alpha_index, :,:], quantity[-1-alpha_index,:,:], quantity.mean(axis=0)
        else:
            return quantity[alpha_index, :,:], quantity[-1-alpha_index, :,:]

class FrozenNumpyEnsembleModel():
    def __init__(self, models) -> None:
        super().__init__()

        self.models = []
        for model in models:
            if isinstance(model, FrozenNumpySurvivalModel):
                self.models.append(model)
            else:
                raise TypeError('One of the entries in `models`'
                                + ' is no a FrozenNumpySurvivalModel')

        self._shape = models[0].forward_out.shape[0]
        for model in models[1:]:
            if model.forward_out.shape[0]!=self._shape:
                raise ValueError('All models must be frozen around the same'+
                                 ' number of features.')            

    def _all_survival_functions(self, T):
        return np.stack(
            [model.survival_function(T).reshape(self._shape,-1)
             for model in self.models])

    def survival_function(self,T):
        return self._all_survival_functions(T).mean(axis=0).squeeze()

    def survival_confidence(self, T, confidence_level=0.95, 
                            return_survival=False):
        S = self._all_survival_functions(T)
        S.sort(axis=0)
        return self._calculate_confidence(
            S,
            confidence_level,
            return_survival)

    def _calculate_confidence(self, quantity, confidence_level, return_mean):

        alpha_0 = .5*(1.0 - (len(self.models)-2)/(max(1, len(self.models)-1)))

        if isinstance(confidence_level, Iterable):
            alpha = .5 * \
                (1.0-np.array(confidence_level))
            alpha_index = (alpha/alpha_0).astype(int)
        else:
            alpha = .5*(1.0-confidence_level)
            alpha_index = int(alpha/alpha_0)

        if return_mean:
            return quantity[alpha_index, :,:], quantity[-1-alpha_index,:,:], quantity.mean(axis=0)
        else:
            return quantity[alpha_index, :,:], quantity[-1-alpha_index, :,:]
        



r"""Energy based models.

In an energy based model, a normalized neural network is used to direcly specify the 
so density. The normalization of the network is done using an
integrator from :mod:`pysare.model.energy_based.integrators`.

The model itself is implemented in :class:`EnergyBasedDensity`.

"""

from .EnergyBasedDensity import EnergyBasedDensity
from . import integrators


from pysare.training._printers import SingleLinePrinter, MultipleLinePrinter
from pysare.training._caching import StateCacher
from torch.nn import Module
import pandas as pd
from numpy.typing import ArrayLike
from time import perf_counter
import string
import os
from pysare.models.SurvivalModel import TorchSurvivalModel
from torch.utils.data import DataLoader
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler
from typing import Optional, Union, Iterable, Callable, Literal, TypeVar, NewType, Any
from os import PathLike
import collections
import time
from types import ModuleType
import matplotlib
import torch
import pysare
import numpy as np
import matplotlib.pyplot as plt

from pysare._typing import Batch, AxisLike
# Log
# ==============================================================================
class TrainingLog(pd.DataFrame):
    r"""
    A training log based on pandas DataFrame returned by the PySaRe training routine. 

    Methods
    -------
    plot(backend):
        Plots the training and validation loss.
    """

    def __init__(self, data):
        super(TrainingLog, self).__init__(data)

    def plot(self, backend: Union[Literal['matplotlib', 'plotext'], AxisLike] = 'matplotlib'):
        r"""
        Plots the training and validation loss using the selected backend.

        Parameters
        ----------
            backend (Union[Literal['plotext','matplotlib'], AxisLike]):
                Backend to be used for plotting. If 'matplotlib' (default)or a 
                matplotlib axis, matploblib is used; and if 'plottex', plotex 
                is used. Any other value will be tried to be used as a matplotlib
                axis, for examle will the module plotext work as input.

        Returns
        ----------
            None
        """
        if (backend is None) or (backend == 'matplotlib'):
            _, ax = plt.subplots(1, 1, clear=True)
            backend = 'matplotlib'
        elif backend == 'plotext':
            import plotext
            ax = plotext
            plotext.clear_data()

        elif isinstance(backend, ModuleType):
            ax = backend
            if backend.__name__.startswith('matplotlib'):
                backend = 'matplotlib'
            else:
                backend = ax.__name__
        elif isinstance(backend, matplotlib.axes._axes.Axes):
            ax = backend
            backend = 'matplotlib'
        else:
            ax = backend
            backend = 'unknown'

        ax.plot(self['epoch'], self['training_loss'], label='Training set')

        if hasattr(ax, 'xlabel'):
            ax.xlabel('Epoch')
            ax.ylabel('Loss value')
        else:
            ax.set_xlabel('Epoch')
            ax.set_ylabel('Loss value')

        if 'validation_loss' in self.columns:
            ax.plot(self['epoch'], self['validation_loss'],
                    label='Validation set')

            min_ind = self['validation_loss'].argmin()

            if min_ind and (backend == 'matplotlib'):
                y_min = self['validation_loss'].iloc[min_ind]
                x_min = self['epoch'].iloc[min_ind]
                ax.plot(x_min, y_min, 'og', label=('Validation minimum: \n  epoch '
                                                   + str(int(x_min)) + '\n  loss     ' +
                                                   format(y_min, '.4f')))

                stop_ind = self['epoch'].argmax()
                y_stop = self['validation_loss'].iloc[stop_ind]
                x_stop = self['epoch'].iloc[stop_ind]
                ax.plot(x_stop, y_stop, 'or', label=('Stopped at: \n  epoch '
                                                     + str(int(x_stop)) + '\n  loss     ' +
                                                     format(y_stop, '.4f')))

        if hasattr(ax, 'legend'):
            ax.legend()
        if hasattr(ax, 'show'):
            ax.show()

# Nonfinite gradient functions
# =============================================================================


def nonfinite_gradient_treatment_none(model):
    return True


def nonfinite_gradient_treatment_skip_step(model):
    for _, param in model.named_parameters():
        if (param.requires_grad) and (not torch.isfinite(param.grad).all()):
            # torch.isnan(param.grad).any() or torch.isinf(param.grad).any():
            return False
    return True


def nonfinite_gradient_treatment_set_zero(model):
    for _, param in model.named_parameters():
        if param.requires_grad:
            param.grad[~torch.isfinite(param.grad)] = 0.
    return True

# Loss function
# =============================================================================


def negative_log_likelihood(model, batch):
    return -model.log_likelihood(*batch)


def loss_aggregation_mean(loss_vector):
    loss = loss_vector.mean()
    return loss, loss.item()*loss_vector.nelement()


def loss_aggregation_sum(loss_vector):
    loss = loss_vector.sum()
    return loss, loss.item()

# Main class


class Trainer():
    r"""
    Trainer that implements a training routine for PyTorch Modules, in particular  
    PyaRe SurvivalModels.

    Examples and more: 
        https://github.com/oholmer/PySaRe

    Parameters
    ----------
        model (Module):
            The model to be trained. Can be any Pytorch Module that is 
            compatible with the parameters loss_function  and data loaders,
            but typically a SurvivalModel.
        optimizer (Optimizer):
            PyTorch Optimizer used to train the model. 
        lr_scheduler (LRScheduler): 
            PyTorch learning rate scheduler. If lr_schedule_after_batch is
            True, rescheduling is done after each batch (default), otherwize 
            rescheduling is done after each epoch. If None (default), no 
            rescheduling is done.
        lr_schedule_after_batch (bool):
            Determines if learning rate scheduling is done after each epoch 
            (if False, default) or after each batch (if True).
        loss_function (Union[Literal['negative_log_likelihood'],Callable):
            Loss function to be used during training. A function that takes
            the model as first arguement and one of the batches that the 
            data loaders iterates over as second argument, and returns the
            vectorized loss in the form of a tensor. The ouput from this 
            function is made into a scalar based on the loss_aggregation
            parameter. If a the literal 'negative_log_likelihood' (default)
            is passed, negative log-likelihood is used as loss function.
        loss_aggregation (Literal['mean', 'sum']):
            Determines how the loss is aggregated into a scalar. If 'mean' 
            (default) the mean is taken, and if 'sum' the sum is taken.
        nonfinite_gradient_treatment (Literal['none', 'skip_step', 'set_zero']):
            Determinies how nonfinite values (NaN or Inf, 
            based on torch.isfinite) in gradients are treated . If 'none',
            a step is taken without any intervention (often leading to a 
            non-functioning model); if 'skip_step' (default), the step is 
            not taken if any gradient contains a nonfinite element; and if 
            'set_zero' any nonfinite element in a gradient is zet zero. 
            Note that Inf in gradients can be handeled using the 
            pre_step_function or by using hooks, which often helps with 
            NaN as well.
        cache_in_memory (bool):
            Determines if the model is saved in memmory or saved to file.
            If False, the model will be saved to file in the file specified by
            best_model_path if provided and otherwise in a temporary file. 
        best_model_path (PathLike):
            Path where the best model parameters are saved, only used if
            chache_in_memmory is Falase and load_best_model=True is passed 
            when running the train method.
        print_frequency (Float):
            Determines minimal time between printouts as 1/print_frequency.
            If None (default), 
        print_new_lines (bool | int):
            If False (default), progress during trainig will be reprinted 
            on the same linte. If True, each printed epoch is printed on a
            new line. If int, a new line will only be printed if the number of
            epochs since last new line is at least the given value. 

    """

    def __init__(self,
                 model: Module,
                 optimizer: Optimizer,
                 lr_scheduler: Optional[LRScheduler] = None,
                 lr_schedule_after_batch: bool = False,
                 loss_function: Union[Literal['negative_log_likelihood'],
                                      Callable[[Module, Batch], torch.Tensor]] = 'negative_log_likelihood',
                 loss_aggregation: Literal['mean', 'sum'] = 'mean',
                 on_nonfinite_gradient: Literal['none',
                                                       'skip_step', 'set_zero'] = 'none',
                 cache_in_memory: bool = False,
                 best_model_path: Optional[PathLike] = None,
                 print_frequency: Optional[float] = None,
                 print_new_lines: Union[bool, int] = False
                 ) -> None:

        self.model = model
        self.optimizer = optimizer

        self.lr_scheduler = lr_scheduler
        self.lr_schedule_after_batch = lr_schedule_after_batch

        self.best_model_cache = StateCacher(
            cache_in_memory, best_model_path, overwrite=True)

        if loss_function == 'negative_log_likelihood':
            self.loss_function = negative_log_likelihood
        else:
            self.loss_function = loss_function

        self.loss_aggregation = loss_aggregation
        if loss_aggregation == 'mean':
            self.loss_aggregation_fun = loss_aggregation_mean
        elif loss_aggregation == 'sum':
            self.loss_aggregation_fun = loss_aggregation_sum
        else:
            raise ValueError("loss_aggregation must be 'mean' or 'sum' ")

        # Printer
        # ======================================================================
        if print_frequency is None:
            print_time = 0.0
        else:
            print_time = 1./print_frequency

        if print_new_lines:
            self.printer = MultipleLinePrinter(
                print_time=print_time, print_new_line_count=int(print_new_lines))
        else:
            self.printer = SingleLinePrinter(print_time=print_time)

        if on_nonfinite_gradient == 'none':
            self.on_nonfinite_gradient_fun = nonfinite_gradient_treatment_none
        elif on_nonfinite_gradient == 'skip_step':
            self.on_nonfinite_gradient_fun = nonfinite_gradient_treatment_skip_step
        elif on_nonfinite_gradient == 'set_zero':
            self.on_nonfinite_gradient_fun = nonfinite_gradient_treatment_set_zero
        else:
            raise ValueError(
                "nonfinite_gradient_treatment must be 'none', 'skip_step', or 'set_zero'.")

        self.log = None

    def train(self,
              num_max_epochs: int,
              training_loader: Iterable[Batch],
              validation_loader: Optional[Iterable[Batch]] = None,
              early_stopping_patience: Optional[int] = None,
              load_best_model: bool = True):
        r"""
        Runs the training routine.

        Examples and more: 
            https://github.com/oholmer/PySaRe

        Parameters
        ----------
            num_max_epochs (int): 
                Maximal number of epochs to train.
            training_loader (Iterable[Batch]): 
                Training data loader that iterates over the batches, 
                typically a PyTorch DataLoader.
            validation_loader (Iterable[Batch]):
                Validation data loader that iterates over the batches, 
                typically a PyTorch DataLoader. If None (default), no 
                validation step is used.
            early_stopping_patience (int):
                Number of epochs without improvement before improvement in 
                validation loss before the training is stopped. If None (default),
                early stopping is not used.
            load_best_model (bool):
                Determines if the best model is to be loaded after training.
                True by default.

        Returns
        ----------
            log (TrainingLog):
                A log (pandas DataFrame) of the training and validation loss for
                each epoch. Use log.plot() to visualize the training.
        """
        # Interpret inputs
        # ======================================================================
        if early_stopping_patience is None:
            early_stopping_patience = num_max_epochs

        if self.lr_scheduler:
            if self.lr_schedule_after_batch:
                lr_schedule_batch_step = self.lr_scheduler.step

                def lr_schedule_epoch_step():
                    return None
            else:
                lr_schedule_epoch_step = self.lr_scheduler.step

                def lr_schedule_batch_step():
                    return None
        else:
            def lr_schedule_epoch_step():
                return None
            def lr_schedule_batch_step():
                return None



        # Setup checkpoint path
        if validation_loader is None:
            load_best_model = False
        else:  
            self.best_model_cache.save(self.model)

        # Main training loop
        # ==========================================================================
        train_loss = []
        validation_loss = []

        best_validation_loss = np.inf
        last_epoch_with_improvement = -1

        stopped_early = False
        for epoch in range(1, num_max_epochs+1):

            self.model.train()

            train_epoch_loss = 0.0

            # Loop over train batches
            # ----------------------------------------------------------------------
            for batch_number, batch in enumerate(training_loader):
                # Compute loss
                loss_vector = self.loss_function(self.model, batch)

                train_batch_loss, loss_vector_sum = self.loss_aggregation_fun(
                    loss_vector)

                train_epoch_loss += loss_vector_sum

                # Backpropagation
                self.optimizer.zero_grad()

                train_batch_loss.backward()

                if self.on_nonfinite_gradient_fun(self.model):
                    self.optimizer.step()
                    lr_schedule_batch_step()

                self.printer.train_update(epoch, batch_number+1,
                                          train_batch_loss.item())

            if self.loss_aggregation == 'mean':
                train_epoch_loss = train_epoch_loss / \
                    len(training_loader.dataset)

            train_loss.append(train_epoch_loss)

            lr_schedule_epoch_step()

            # Loop over validation batches
            # ----------------------------------------------------------------------
            if validation_loader:
                self.model.eval()
                validation_epoch_loss = 0.
                with torch.no_grad():
                    for batch in validation_loader:

                        validation_batch_loss_vector = self.loss_function(
                            self.model, batch)
                        _, loss_vector_sum = self.loss_aggregation_fun(
                            validation_batch_loss_vector)
                        validation_epoch_loss += loss_vector_sum

                    if self.loss_aggregation == 'mean':
                        validation_epoch_loss /= len(validation_loader.dataset)

                validation_loss.append(validation_epoch_loss)

                # Early stopping and checkpoint
                # ----------------------------------------------------------------------
                if validation_epoch_loss < best_validation_loss:
                    best_validation_loss = validation_epoch_loss
                    last_epoch_with_improvement = epoch
                    if load_best_model:
                        self.best_model_cache.save(self.model)
                if epoch - last_epoch_with_improvement > early_stopping_patience:
                    stopped_early = True
                    break

            else:
                validation_epoch_loss = None

            self.printer.finalize_update(
                epoch, batch_number+1, train_epoch_loss, validation_epoch_loss, train_batch_loss)

        # ==========================================================================
        # Finalize

        # Test if a checpoints has been saved
        model_restored_from_checkpoint = False
        if load_best_model and (not self.best_model_cache.empty):
            self.best_model_cache.load(self.model)

        self.best_model_cache.delete_temp()

        self.printer.finalize()

        if stopped_early:
            print(
                f"Training stoped at: \n   Epoch: {epoch:.0f}  (no improvement for {early_stopping_patience:.0f} epochs)")
        else:
            print(
                f"Training stoped at:\n   Epoch: {epoch:.0f} (maximal number)")

        if model_restored_from_checkpoint:
            print(
                f"Restored model parameters from:\n   Epoch: {last_epoch_with_improvement:.0f}")

        print(f"   Validation loss: {best_validation_loss:.5f}")

        if validation_loader:
            self.log = TrainingLog(data={'epoch': range(1, epoch+1),
                                         'training_loss': train_loss,
                                         'validation_loss': validation_loss})
        else:
            self.log = TrainingLog(data={'epoch': range(1, epoch+1),
                                         'training_loss': train_loss})
        return self.log

    def plot(self, backend: Union[Literal['matplotlib', 'plotext'], AxisLike] = 'matplotlib'):
        r"""
        Plots the training and validation loss using the selected backend.

        Parameters
        ----------
            backend (Union[Literal['plotext','matplotlib'], AxisLike]):
                Backend to be used for plotting. If 'matplotlib' (default)or a 
                matplotlib axis, matploblib is used; and if 'plottex', plotex 
                is used. Any other value will be tried to be used as a matplotlib
                axis, for example will the module plotext work as input.

        Returns
        ----------
            None
        """
        if self.log is not None:
            return self.log.plot(backend)
        else:
            raise ValueError('Training must be run before plotting log.')


"""A sub-package for training PySaRe models and other PyTorch models.

The following classes are available:

- :class:`Trainer` -- Used to train models
- :class:`LRFinder`     -- Used to find an appropriate learning rate
- :class:`GridSearch` --  Used for hyperparameter searches
"""

from ._LRFinder import LRFinder
from ._Trainer import Trainer
from ._GridSearch import GridSearch




import pandas as pd
import numpy as np 
from tqdm.autonotebook import tqdm

class GridSearch():
    r"""Creates and executes grid searches.

    
    """
    def __init__(self, jobs:pd.DataFrame = None, parameters: dict = None) -> None:

        if isinstance(jobs, pd.DataFrame):
            self.jobs = jobs
        elif jobs is None:
            self.jobs = None
        else:
            raise ValueError("Argument jobs must be either DataFrame or None")
        
        if isinstance(parameters, dict):
            for key, values in parameters.items():
                self.add_parameter(key, values)
        elif parameters is not None:
            raise ValueError("Argument parameters must be either dict or None")
        
        self.result = None

    def add_parameter(self, key, values):
        r"""Adds a parameter/column to the list of jobs.
        
        
        """
        if self.jobs is None:
            self.jobs = pd.DataFrame(data={key: values})
        else:
            if np.isscalar(values[0]):
                new_col =  np.array(values).repeat(len(self.jobs), axis=0)
            else:
                new_col = list(values)
                new_col.append([])
                new_col =  np.array(new_col,dtype=object)[:-1].repeat(len(self.jobs), axis=0)
            self.jobs = pd.concat([self.jobs]*len(values), ignore_index=True)
            self.jobs[key] = new_col

    # def add_generator(self, key, generator):
    #     pass
    # def add_random(self, key, data, method):
    #     pass
    
    # def generate(self, num_rows, num_blocks):
    #     pass

    def run_jobs(self,experiment, suppress_printing=False):
        r"""Runs the grid search.
        
        Parameters
        ----------
        
        experiment : callable
            A function that takes one job (row in jobs) as input and outputs 
            one or two dicts. The first dict is aggregated in the main process 
            and included in the return results and the second dict is only 
            saved to file.
        """
        res = []

        
        for ind, job in tqdm(self.jobs.iterrows(), total=self.jobs.shape[0]):
            # with nostdout(supress_printing):
                res.append(experiment(**job))

        self.result = pd.concat((self.jobs,pd.DataFrame.from_records(res)), axis=1)
        return self.result


from torch.nn import Module
import torch

class MLP(Module):
    def __init__(self, input_size, output_size, hidden_sizes,
                 activation=torch.nn.ReLU, dropout=0.0, batch_norm=False):
        super().__init__()

        self.input_size = input_size

        layer_input_size = input_size
        layerlist = []
        for layer_output_size in hidden_sizes:
            layerlist.append(torch.nn.Linear(
                layer_input_size, layer_output_size))
            layerlist.append(activation())
            if batch_norm:
                layerlist.append(torch.nn.BatchNorm1d(layer_output_size))
            if dropout:
                layerlist.append(torch.nn.Dropout(dropout))

            layer_input_size = layer_output_size

        layerlist.append(torch.nn.Linear(layer_output_size, output_size))

        self.layers = torch.nn.Sequential(*layerlist)

    def forward(self, X):
        logits = self.layers(X)

        return logits



import numpy as np
import matplotlib.pyplot as plt
import torch.utils.data
import pysare
import numpy as np
import matplotlib.pyplot as plt
import lifelines as lf
import pandas as pd
from pandas import DataFrame
from collections.abc import Iterable
from matplotlib.axes._axes import Axes
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models._numpy.NumpySurvivalModel import FrozenNumpySurvivalModel
from pysare.models._numpy.NumpySurvivalModel import NumpySurvivalModel
from pysare.data import Dataset
from torch.utils.data.dataloader import DataLoader
from typing import Union, Optional, Tuple, Literal
from numpy.typing import ArrayLike


from pysare._typing import X, T, E


def concordance_index(model: Union[TorchSurvivalModel, NumpySurvivalModel],
                      data: Union[Tuple[X, T, E], Dataset, DataLoader],
                      censoring: Union[Literal['KM', 'none'],
                                       FrozenNumpySurvivalModel] = 'KM',
                      plot: Union[bool, Axes, Iterable[Axes]] = False) -> DataFrame:
    r"""Concordance index of a model on a dataset.

    Calculates the censoring-corrected truncated time-dependent 
    concordance index (C-index), where: 

    - Time-dependent refers to that no PH assumption is made and instead the 
      comparison is done based on the value of the predicted survival function 
      at the time of one of the event times that are compared. 
    - Censoring-corrected refers to that inverse probability weights are used to 
      make the result less dependent on the censoring distribution
    - Truncated refers to that the result is truncated up to a specific time so 
      that only data points with event times before this time are compared (the 
      value at the largest truncation time corresponds to the conventional index). 

    See references [1] and  [2] for more information.

    Parameters
    ----------

    model : SurvivalModel or NumpySurvivalModel:
        Model to be evaluated. If data is a tuple (X, T, E) of non-tensors, 
        the model should be a NumpySurvivalModel, and otherwise it should 
        be a SurvivalModel.
    data : tuple(X,T,E) or Dataset or Dataloader:
        The data to be used for evaluation. If data is a tuple (X, T, E) of 
        non-tensors, the model should be a NumpySurvivalModel, and otherwise 
        it should be a SurvivalModel.
    censoring :  'KM' or 'none' or FrozenNumpySurvivalModel, default='KM'
        Determines how the censoring distribution is calculated. If 'KM' 
        it is estimated using the Kaplan-Meier estimate of the data, if 
        'None' no censoring is use, and if it is a FrozenNumpySurvivalModel
        this model is used.
    plot : bool or Axis or Iterable[Axes], default=False
        If True, a plot of the result (index and censoring) will be created 
        in a new figure; if one or two Axes are passed, the plots will be 
        made in these.

    Returns
    ----------

        result : DataFrame
            A pandas DataFrame with the result.

    References
    ----------

    [1] Antolini, L., Boracchi, P. and Biganzoli, E. (2005), A time-dependent 
    discrimination index for survival data. Statist. 
    https://doi.org/10.1002/sim.2427

    [2] Uno, H., Cai, T., Pencina, M.J., D'Agostino, R.B. and Wei, L.J. (2011), 
    On the C-statistics for evaluating overall adequacy of risk prediction 
    procedures with censored survival data. Statist. Med., 30: 1105-1117. 
    https://doi.org/10.1002/sim.4154

    """

    if hasattr(model, 'eval'):
        model.eval()

    # Interpret inputs
    # ==========================================================================
    if type(data) is tuple:  # data is a tuple (X, T, E)
        dataloader = (data,)
    elif isinstance(data, Iterable):  # data is a DataLoader
        dataloader = data
    else:  # Data is a Dataset
        dataloader = (data[:],)

    dataset = data

    if isinstance(model,pysare.models.TorchSurvivalModel):
        def get_time_vector(time, length):
            return time.repeat(length)
    else:   
        def get_time_vector(time, length):
            return time.repeat(length).reshape(-1,1)

    # Full vector of times  and events
    # ===========================================================================
    T = []
    E = []

    for X_batch, T_batch, E_batch in dataloader:
        T.append(T_batch.reshape(-1,))
        E.append(E_batch.reshape(-1,))

    if torch.is_tensor(T[0]):
        T = torch.concat(T)
        E = torch.concat(E)
        T_np = np.array(T)
        E_np = np.array(E)
    else:
        T = np.concatenate(T)
        E = np.concatenate(E)
        T_np = T
        E_np = E

    # Event times (not censorings)
    T_event = T_np[E_np]

    # Uniqe times between the first and last recorded event
    tau = np.sort(
        np.unique(T_np[(T_np <= T_event.max()) & (T_np >= T_event.min())]))

    # Censoring
    # =========================================================================
    if censoring == "KM":  # Censoring estimated using Kaplan-Meier
        KMC = lf.KaplanMeierFitter()
        KMC.fit(T_np, ~E_np)
        C_T = KMC.survival_function_at_times(T_np).values.reshape(-1,)
        C_tau = KMC.survival_function_at_times(tau).values.reshape(-1,)
    elif (censoring == 'none') or (censoring is False):  # No censoring / censoring = 1
        C_T = np.ones((T_np.shape[0],))
        C_tau = np.ones((tau.shape[0],))
    else:  # censoring is a FrozenSurvivalModel
        C_T = censoring.survival_function(T).reshape(-1,)
        C_tau = censoring.survival_function(tau).reshape(-1,)

    # =========================================================================

    A = np.zeros_like(T_np)
    B = np.zeros_like(T_np)
    with torch.no_grad():
        for n in range(T.shape[0]):

            S_Tn = []
            for X_batch, _, _ in dataloader:
                # with torch.no_grad():
                S_Tn.append(np.array(model.survival_function(
                    X_batch, get_time_vector(T[n:n+1],X_batch.shape[0])).flatten(), copy=False))

            # S_Tn = [np.array(model.survival_probability(X_batch,
            #                                             T[n:n+1].repeat(X_batch.shape[0])),
            #                  copy=False)
            #         for X_batch, _, _ in dataloader]
            S_Tn = np.concatenate(S_Tn)

            comp = (T_np[n] < T_np)  # .reshape(-1,)

            A[n] = ((S_Tn[n] < S_Tn) & comp).sum()
            B[n] = comp.sum()

            # A[n] = ((S_Tn[n] < S_Tn) & comp).sum()/(C_T[n]**2)*E_np[n]
            # B[n] = comp.sum()/(C_T[n]**2)*E_np[n]

    A /= (C_T**2)
    A[~E_np] = 0.
    B /= (C_T**2)
    B[~E_np] = 0.

    C = np.zeros(tau.shape)

    for n in range(tau.shape[0]):
        ind = T_np <= tau[n]
        C[n] = A[ind].sum()/B[ind].sum()

    # Plot
    if plot is not False:
        if plot is True:
            if censoring is not False:
                fig, ax = plt.subplots(nrows=2)
            else:
                fig, ax = plt.subplots(nrows=1)
                ax = (ax,)

        elif isinstance(plot, Iterable):
            ax = tuple(plot)
        else:
            ax = (plot,)

        if (censoring is not False) and (len(ax) > 1):
            if censoring == 'KM':
                KMC.plot_survival_function(ax=ax[1])
                ax[1].legend(['Kaplan-Meier Estimate',
                              'Confidence Interval (95%)'])
            else:
                ax[1].plot(T, C_T)

            ax[1].set_xlabel('Time')
            ax[1].set_ylabel('Censoring Distribution')

        ax[0].plot(tau, C)
        ax[0].set_ylabel('Truncated C-index')
        ax[0].set_xlabel('Truncation time')

        if plot is True:
            plt.tight_layout()

    return pd.DataFrame(data={'C_index': C, 'truncation_time': tau, 'censoring_survival': C_tau})



r"""Mixture models

Models based on a mixture of multiple components. The parameters of each 
component and the weights in the mixture are determined based on the 
output from the forward function, see :ref:`Forward output <forward-output>`
below.

The following models are available:

- :class:`WeibullMixture`
    A mix of Weibull distributions. Two parameters per component,
    shape and scale.
- :class:`GaussianMixture`
    A mix of Gaussian/Normal distributions, reflected around zero. 
    Two parameters per component, location and variance.

.. _forward-output:

Forward output
--------------

Let N be the number of components in the mixture and M the number of 
parameters for each component. A row of the output from forward should 
have the format `[p_1_1, ..., p_1_M, p_2_1, ...,P_N_M, w_1, ..., w_1]`
where p_i_j is the value of parameter j in component i, before activation;
and w_1,..., w_n are the mixing weights, before softmax. The size of the 
forward output should therefore be (batch_size, (M+1)*N).

"""

from . import _mixers
from .GaussianMixture import GaussianMixture
from .WeibullMixture import WeibullMixture


from numpy import linspace
import torch
from pysare.models.distributions import Uniform, Weibull
from pysare.data.Dataset import Dataset


class CategoricalWeibull(Weibull):
    def __init__(self, shape_categories, scale_categories) -> None:
        super().__init__()

        self.shape_categories = torch.nn.Parameter(
            shape_categories, requires_grad=False)
        self.scale_categories = torch.nn.Parameter(
            scale_categories, requires_grad=False)

    def forward(self, X):
        return torch.concat((self.shape_categories[X[:, [0]]],
                             self.scale_categories[X[:, [1]]]), 1)


def seed_to_generator(seed):

    if isinstance(seed, torch.Generator):
        return seed
    elif seed is None:
        return None
    else:
        generator = torch.Generator()
        generator.manual_seed(seed)
        return generator


class SimulatedCategoricalWeibull():
    def __init__(self,
                 shuffle_categories=False,
                 shape_categories=linspace(1.5, 5.0, 10),
                 scale_categories=linspace(1, 3.0, 10),
                 max_censoring_time=3.0,
                 seed=None,
                 dtype=torch.get_default_dtype()) -> None:

        generator = seed_to_generator(seed)

        self.shape_categories = torch.tensor(shape_categories, dtype=dtype)
        self.scale_categories = torch.tensor(scale_categories, dtype=dtype)

        self.num_categories = {'shape': len(self.shape_categories),
                               'scale': len(self.scale_categories)}
        
        self.max_censoring_time = max_censoring_time

        if shuffle_categories:
            self.shape_categories = self.shape_categories[
                torch.randperm(len(self.shape_categories),
                               generator=generator)]
            self.shape_categories = self.scale_categories[
                torch.randperm(len(self.scale_categories),
                               generator=generator)]

        float_tensor = torch.empty(0, dtype=dtype)
        int_tensor = torch.empty(0, dtype=torch.int64)

        self.event_distribution = CategoricalWeibull(
            shape_categories=self.shape_categories,
            scale_categories=self.scale_categories).to_numpy(
                data=(int_tensor, float_tensor, float_tensor))

        self.censoring_distribution = Uniform().to_numpy(
            data=(float_tensor,
                  float_tensor,
                  float_tensor)).freeze([[0.0, max_censoring_time]])

    def sample_features(self, num_samples, seed=None):

        generator = seed_to_generator(seed)

        return torch.concat((torch.randint(0, len(self.shape_categories),
                                           (num_samples, 1),
                                           generator=generator),
                             torch.randint(0, len(self.scale_categories),
                                           (num_samples, 1),
                                           generator=generator)), 1)

    def generate(self, num_samples, seed=None):
        generator = seed_to_generator(seed)
        X = self.sample_features(num_samples, seed=generator)
        T_event = self.event_distribution.sample(X, num_samples=1,
                                                 generator=generator).squeeze()
        T_cens = self.censoring_distribution.sample(num_samples=num_samples,
                                                    generator=generator).squeeze()
        E = T_event <= T_cens
        T = T_event
        T[~E] = T_cens[~E]
        return Dataset(X, T, E, copy=False)



import numpy
import pandas


def _simulate_finite_horizon(failure_distribution, horizon, policy):
    N_f = 0
    N_p = 0

    h = horizon

    while True:

        T = failure_distribution.sample()

        r = policy.replacement_age(h)
        if T < r:
            h -= T
            if h <= 0.0:
                return N_p, N_f
            N_f += 1
        else:
            h -= r
            if h <= 0.0:
                return N_p, N_f
            N_p += 1


def simulate(failure_distribution, horizon, c_p, c_f, policy, num_simulations=1):
    N_p = numpy.zeros(num_simulations)
    N_f = numpy.zeros(num_simulations)

    try:
        failure_distribution.sample()
    except Exception as error:
        raise Exception("Was no able to sample failure_distribuiton. "
                    "Try converting it to a spline  approximation using"
                    " to_spline_approximation.") from error
        

    for n in range(num_simulations):
        N_p[n], N_f[n] = _simulate_finite_horizon(
            failure_distribution, horizon, policy)

    return pandas.DataFrame({'num_failures': N_f, 'num_preventive': N_p, 'cost': N_f*c_f+N_p*c_p})



import torch
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models.piecewise.grid import DiscretizationGrid
from pysare.models.utils import MLP
from abc import abstractmethod


def reverse_cumsum(x):
    cum_sum = torch.cumsum(x, dim=1)
    return x - cum_sum + cum_sum[:, -1].view(-1, 1)


class LinearDensity(TorchSurvivalModel):
    r"""Implements an piecewise linear density model.
    
    The forward output size should be 
        number of intervals + 2 + potential extra inputs required by the grid

    See :mod:`pysare.models.piecewise` for information on how to use 
    piecewise models.
    """
    def __init__(self, grid: DiscretizationGrid):
        super(LinearDensity, self).__init__()
        self.grid = grid

    def _bin_and_normalize(self, z, T):
        # Number of subjects
        N = T.shape[0]
        # Binwidths


        disc, z = self.grid.bin(T, z)

        e = torch.exp(z)

        # Calculate S times Z at each tau
        # dF = e[:, :-1]*disc.dt

        dF = e[:, 1:].clone()
        dF[:, :-1] += e[:, :-2]
        dF[:, :-1] *= disc.dt/2

        SZ = reverse_cumsum(dF)

        # Normalization constant is first collumn in S*Z (1*Z)
        Z = SZ[:, 0]

        e_ind = e[range(N), disc.ind]
        e_next_ind = e[range(N), disc.ind+1]

        return SZ, Z, z, e, disc, e_ind, e_next_ind

    def forward_output_size(self):
        return self.grid.num_intervals()+self.grid.extra_forward_output_size()+2

    def max_time(self):
        return self.grid.max_time()

    def _log_likelihood(self, z, T, E):
        N = T.shape[0]

        SZ, Z, z, e, disc, e_ind, e_next_ind = self._bin_and_normalize(z, T)

        log_likelihood = -torch.log(Z)

        ind_f = E
        ind_c = ~E

        log_likelihood[ind_f] += torch.log(e[ind_f, disc.ind[ind_f]] + (T[ind_f]-disc.tau_ind[ind_f]) * (
            e[ind_f, disc.ind[ind_f]+1]-e[ind_f, disc.ind[ind_f]])/disc.dt_ind.expand(N)[ind_f])
        log_likelihood[ind_c] += torch.log(SZ[ind_c, disc.ind[ind_c]+1] + e_next_ind[ind_c]*(disc.tau_next_ind[ind_c] - T[ind_c]) + (
            (disc.tau_next_ind[ind_c] - T[ind_c])**2)/(2*disc.dt_ind.expand(N)[ind_c])*(e_ind[ind_c] - e_next_ind[ind_c]))

        return log_likelihood

    def _density_function(self, z, T):
        SZ, Z, z, e, disc, e_ind, e_next_ind = self._bin_and_normalize(z, T)

        N = T.shape[0]
        f = (e[range(N), disc.ind] + (T-disc.tau_ind) *
             (e[range(N), disc.ind+1]-e[range(N), disc.ind])/disc.dt_ind)/Z
        return f

    def _survival_function(self, z, T):
        SZ, Z, z, e, disc, e_ind, e_next_ind = self._bin_and_normalize(z, T)
        N = T.shape[0]

        S = (SZ[range(N), disc.ind+1] + e_next_ind*(disc.tau_next_ind - T) +
             ((disc.tau_next_ind - T)**2)/(2*disc.dt_ind)*(e_ind - e_next_ind))/Z
        return S

    def forward(self, X):
        return X

    @classmethod
    def MLP_implementation(cls, grid, input_size, hidden_sizes,
                           activation=torch.nn.ReLU, dropout=False,
                           batch_norm=False):
        return MLPLinearDensity(grid, input_size, hidden_sizes,
                                activation=torch.nn.ReLU, dropout=0.0,
                                batch_norm=False)

class MLPLinearDensity(LinearDensity):
    def __init__(self, grid, input_size, hidden_sizes,
                 activation=torch.nn.ReLU, dropout=0.0, batch_norm=False):
        super(MLPLinearDensity, self).__init__(grid=grid)

        self.input_size = input_size

        self.layers = MLP(input_size, grid.num_intervals() + grid.extra_forward_output_size() + 2,
                          hidden_sizes, activation, dropout, batch_norm)

    def forward(self, X):
        logits = self.layers(X)

        return logits




import torch
from pysare.models.SurvivalModel import TorchSurvivalModel
from pysare.models.piecewise.grid import DiscretizationGrid
from pysare.models.utils.MLP import MLP


def reverse_cumsum(x):
    cum_sum = torch.cumsum(x, dim=1)
    return x - cum_sum + cum_sum[:, -1].view(-1, 1)


class ConstantDensity(TorchSurvivalModel):
    r"""Implements an piecewise constant density model
    
    The forward output size should be 
        number of intervals + 1 + potential extra inputs required by the grid

    See :mod:`pysare.models.piecewise` for information on how to use 
    piecewise models.
    """

    def __init__(self, grid: DiscretizationGrid):
        super(ConstantDensity, self).__init__()
        self.grid = grid

    def _bin_and_normalize(self, z, T):

        # Number of subjects
        N = T.shape[0]
        # Binwidths

        disc, z = self.grid.bin(T, z)

        e = torch.exp(z)

        # Calculate S times Z at each tau
        dF = e.clone()
        dF[:, :-1] *= disc.dt
        SZ = reverse_cumsum(dF)

        # Normalization constant is first collumn in S*Z (1*Z)
        Z = SZ[:, 0]

        return SZ, Z, z, e, disc.ind, disc.tau_next_ind

    def forward_output_size(self):
        return self.grid.num_intervals()+self.grid.extra_forward_output_size()+1

    def max_time(self):
        return self.grid.max_time()

    def _log_likelihood(self, z, T, E):

        SZ, Z, z, e, ind, tau_next = self._bin_and_normalize(z, T)

        log_likelihood = -torch.log(Z)

        ind_f = torch.where(E)[0]
        ind_c = torch.where(~E)[0]
        log_likelihood[ind_f] += z[ind_f, ind[ind_f]]
        log_likelihood[ind_c] += torch.log(SZ[ind_c, ind[ind_c]+1] +
                                           (tau_next[ind_c]-T[ind_c])*e[ind_c, ind[ind_c]])

        return log_likelihood

    def _density_function(self, z, T):
        SZ, Z, z, e, ind, tau_next = self._bin_and_normalize(z, T)
        N = T.shape[0]
        f = e[range(N), ind]/Z
        return f

    def _survival_function(self, z, T):
        SZ, Z, z, e, ind, tau_next = self._bin_and_normalize(z, T)
        N = T.shape[0]
        S = (SZ[range(N), ind+1] + (tau_next-T)*e[range(N), ind])/Z
        return S

    def forward(self, X):
        return X

    @classmethod
    def MLP_implementation(cls, grid, input_size, hidden_sizes,
                           activation=torch.nn.ReLU, dropout=False,
                           batch_norm=False):
        return MLPConstantDensity(grid, input_size, hidden_sizes,
                                  activation, dropout,
                                  batch_norm)


class MLPConstantDensity(ConstantDensity):
    def __init__(self, grid, input_size, hidden_sizes,
                 activation=torch.nn.ReLU, dropout=0.0, batch_norm=False):
        super(MLPConstantDensity, self).__init__(grid=grid)

        self.input_size = input_size

        self.layers = MLP(input_size, grid.num_intervals() + grid.extra_forward_output_size() + 1,
                          hidden_sizes, activation, dropout, batch_norm)

    def forward(self, X):
        return self.layers(X)



import numpy as np
from pysare.data import Dataset
import matplotlib.pyplot as plt


class WeibullUniformParameters(Dataset):
    def __init__(self, N=300, t_m=3, scale_range=[1, 3], shape_range=[0.5, 5]):
        

        X, T, E = simulate_weibull_uniform_parameters(N, t_m, scale_range, shape_range)
        
        super(WeibullUniformParameters, self).__init__(X, T, E)
        self.num_features = 2

        self.scale_range = scale_range
        self.shape_range = shape_range
        self.t_m = t_m

    def compare_survival_function(self, models, num_scale, num_shape, num_times=100, ax=None):
        models = np.array(models).reshape(-1,)
        t = np.linspace(0, self.t_m, num_times)

        fig = None
        if ax is None:
            fig, ax = plt.subplots(num_scale, num_shape, clear=True)

        # Make sure ax is two dimensional array
        ax = np.array(ax)
        ax = ax.reshape(num_scale, num_shape)

        for n, la in enumerate(np.linspace(self.scale_range[0],
                                           self.scale_range[1],
                                           num_scale)):
            for m, k in enumerate(np.linspace(self.shape_range[0],
                                              self.shape_range[1],
                                              num_shape)):

                ax[n, m].plot(t, np.exp(-(t/la)**k))
                for model in models:
                    ax[n, m].plot(t, model.survival_probability([[k, la]], t))
                ax[n, m].set_xlim([0, self.t_m])
                ax[n, m].set_ylim([0, 1])
                ax[n,m].set_xlabel('t')
                ax[n,m].set_ylabel('S(t)')

        ax[0,-1].legend(['True','Model'])
        plt.tight_layout()
        return fig, ax

    def compare_model(self, model, scale_parameter, shape_parameter, num_times=100, ax=None):

        t = np.linspace(0, self.t_m, num_times)

        if not ax:
            fig, ax = plt.subplots(2, 1, clear=True)

        ax[0].plot(t, np.exp(-(t/scale_parameter)**shape_parameter))
        ax[0].plot(t, model.survival_probability(
            [[shape_parameter, scale_parameter]], t))

        ax[1].plot(t, shape_parameter/scale_parameter
                   * (t/scale_parameter)**(shape_parameter-1)
                   * np.exp(-(t/scale_parameter)**shape_parameter))
        ax[1].plot(t, model.lifetime_density(
            [[shape_parameter, scale_parameter]], t))

def simulate_weibull_uniform_parameters(N, t_m=3, scale_range=[1, 3], shape_range=[0.5, 5]):
    X1 = np.random.uniform(
        low=shape_range[0], high=shape_range[1], size=[N, 1])
    X2 = np.random.uniform(
        low=scale_range[0], high=scale_range[1], size=[N, 1])
    X = np.concatenate((X1, X2), axis=1)

    def failure_dist(x): return np.random.weibull(x[0]) * x[1]
    def censor_dist(x): return np.random.uniform(low=0, high=t_m)

    T, E = simulate(failure_dist, censor_dist, X)

    return X, T, E


def simulate(FailureDist, CensorDist, V):
    T = np.zeros((V.shape[0], 1))
    C = np.zeros_like(T)

    for n in range(V.shape[0]):
        T[n] = FailureDist(V[n])
        C[n] = CensorDist(V[n])

    E = T <= C
    T = np.concatenate((T, C), axis=1).min(axis=1).reshape(-1,)

    return T, E.reshape(-1,)



"""Discretization grids to be used with piecewise models.

The following grids are available:

- :class:`EquidistantGrid`
    Defines a fixed equidistant grid.
- :class:`FixedGrid`
    Defines a fixed grid based on a supplied list of grid points.
- :class:`TrainableGrid`
    Defines a grid where the grid points are trainable. 
- :class:`FixedGrid`
    Defines a grid where the grid points are determined based on the 
    output from the models forward function.
    
A custom grid can be defined by subclassing the base class :class:`DiscretizationGrid`.

.. autosummary:: 
    :toctree: generated
    :recursive:

    EquidistantGrid
    IndividualizedGrid
    FixedGrid
    TrainableGrid    
    DiscretizationGrid

"""

from ._grid import EquidistantGrid
from ._grid import DiscretizationGrid
from ._grid import IndividualizedGrid
from ._grid import FixedGrid
from ._grid import TrainableGrid


from abc import abstractmethod
from dataclasses import dataclass
import torch
from torch import Tensor
from typing import Any


@dataclass
class Discretization:
    ind: Tensor
    dt: Tensor
    dt_ind: Tensor
    tau_ind: Tensor
    tau_next_ind: Tensor


class DiscretizationGrid(torch.nn.Module):
    r"""Superclass for discretization grids.

    Superclass this class to define a custom grid.    
    """

    @abstractmethod
    def bin(self, t, z):
        pass

    def max_time(self):
        r"""Returns the maximal time in the discretisation."""
        pass

    def num_intervals(self):
        r"""Returns the number of intervals in the grid."""
        pass

    def extra_forward_output_size(self):
        r"""Returns the number elements required by the grid, in adition to 
        those used by the piecewise model."""
        pass


class EquidistantGrid(DiscretizationGrid):
    r"""Implements an uniform grid.

    A fixed grid with equidistant grid points that can not be trained.

    Parameters
    ----------

    max_time : float>0
        The last point in the discretization.
    num_intervals : int>0
        An integer specifying the number of intervals in the discretization.

    """

    def __init__(self, max_time: float, num_intervals: int):
        super().__init__()

        # num_intervals is and int and is handled using bufffer to remain so
        self._num_intervals = torch.nn.Parameter(torch.tensor(
            num_intervals).type(torch.long), requires_grad=False)

        # dt is a float and is handled as a Parameter so that it changes type with model
        self.dt = torch.nn.Parameter(torch.tensor(max_time/num_intervals,
                                                  dtype=torch.get_default_dtype()),
                                     requires_grad=False)

    def extra_forward_output_size(self):
        return 0

    def num_intervals(self):
        return self._num_intervals

    def to(self, *args, **kwargs):
        rval = super().to(*args, **kwargs)

        # N should remain long
        self._num_intervals = self._num_intervals.to(dtype=torch.long)
        return rval

    def max_time(self):
        return (self._num_intervals*self.dt).item()

    def bin(self, t, z):

        ind = torch.clip(torch.floor_divide(t, self.dt).to(
            torch.long), min=None, max=self._num_intervals-1)

        return Discretization(
            ind=ind,
            dt=self.dt,
            dt_ind=self.dt,
            tau_ind=self.dt*ind,
            tau_next_ind=self.dt*(ind+1)), z


class FixedGrid(DiscretizationGrid):
    r"""Implements a fixed grid.

    A fixed grid where the discretization is provided by the user, and can not 
    be trained.

    Parameters
    ----------
    discretization_times : (num_intervals+1,) array-like of increasing floats
        Discretization times. Will be converted to a tensor with default dtype 
        and device. The times must be increasing, no test of this is done.
    """

    def __init__(self, discretization_times: Any):
        super().__init__()

        self.discretization_times = torch.nn.Parameter(torch.tensor(
            discretization_times,
            dtype=torch.get_default_dtype()), requires_grad=False)
        self._dt = torch.nn.Parameter(torch.diff(self.discretization_times),
                                      requires_grad=False)

    def extra_forward_output_size(self):
        return 0

    def to(self, *args, **kwargs):
        rval = super().to(*args, **kwargs)
        return rval

    def max_time(self):
        return self.discretization_times[-1].item()

    def num_intervals(self):
        return self.discretization_times.shape[0]-1

    def bin(self, t, z):

        N = t.shape[0]

        ind = torch.searchsorted(
            self.discretization_times, t,
            right=True).clip(None,
                             self.discretization_times.shape[0]-1)

        tau_next_ind = self.discretization_times[ind]
        ind -= 1
        tau_ind = self.discretization_times[ind]

        # ind = torch.clip(torch.floor_divide(t, self.dt).to(
        # torch.long), min=None, max=self.N-1)

        return Discretization(
            ind=ind,
            dt=self._dt,
            dt_ind=self._dt[ind],
            tau_ind=tau_ind,
            tau_next_ind=tau_next_ind), z


class TrainableGrid(DiscretizationGrid):
    r"""Implements a trainable grid.

    A trainable grid where the discretization is parameterized by a tensor 
    which, after applying softmax and normalizing so that they sum up to the
    maximal time, specifies the size of each interval.

    Parameters
    ----------
    discretization_times : (num_intervals+1,) array-like of increasing floats
        Discretization times. Will be used to determine the maximal time (last
        value) and the initial parameterization of the grid.
    """

    def __init__(self, discretization_times: Any):
        super().__init__()

        self._max_time = float(discretization_times[-1])

        log_prop_dt = torch.log(torch.diff(
            torch.tensor(discretization_times,
                         dtype=torch.get_default_dtype())))

        self.log_prop_dt = torch.nn.Parameter(
            log_prop_dt-torch.max(log_prop_dt),
            requires_grad=True)

    def extra_forward_output_size(self):
        return 0

    def to(self, *args, **kwargs):
        rval = super().to(*args, **kwargs)
        return rval

    def max_time(self):
        return self._max_time

    def num_intervals(self):
        return self.log_prop_dt.shape[0]

    def discretization_times(self):
        delta = torch.softmax(self.log_prop_dt, 0)
        dt = self._max_time*delta
        tau = dt.cumsum(0)

        return torch.concat((torch.zeros((1,), dtype=tau.dtype,
                                         device=tau.device),
                             tau))

    def bin(self, t, z):

        N = t.shape[0]

        delta = torch.softmax(self.log_prop_dt, 0)

        dt = self._max_time*delta

        tau = dt.cumsum(0)

        ind = torch.searchsorted(
            tau, t, right=False).clip(None, tau.shape[0]-1)

        tau_next_ind = tau[ind]
        tau_ind = tau_next_ind - dt[ind]

        return Discretization(
            ind=ind,
            dt=dt,
            dt_ind=dt[ind],
            tau_ind=tau_ind,
            tau_next_ind=tau_next_ind), z


class IndividualizedGrid(DiscretizationGrid):
    r"""Implements a grid where the discretization depends on the features.

    This grid uses the output from the forward function of the model to 
    determine the discretization individualy for each sampel.

    To use this grid, the forward functoin of the model shold return 
    num_intervals extra entries, positioned last in the tensor. After applying 
    softmax, theses are used to determine the size of each interval so that 
    they sum up to max_time. 

    Parameters
    ----------
    max_time : float>0
        The last point in the discretization.
    num_intervals : int>0
        An integer specifying the number of intervals in the discretization.
    """

    def __init__(self, max_time: float, num_intervals: int):
        super().__init__()

        # num_intervals is and int and is handled using bufffer to remain so
        self._num_intervals = torch.nn.Parameter(torch.tensor(
            num_intervals).type(torch.long), requires_grad=False)

        self._max_time = float(max_time)

    def extra_forward_output_size(self):
        return self._num_intervals

    def max_time(self):
        return self._max_time

    def num_intervals(self):
        return self._num_intervals

    def bin(self, t, z):

        # The last self.num_intervals entries in z (dim 1) should be used here
        v = z[:, -self._num_intervals:]
        z = z[:, :-self._num_intervals]

        N = t.shape[0]

        delta = torch.softmax(v, 1)

        dt = self._max_time*delta
        tau = dt.cumsum(1)

        ind = (tau[:, :-1] < t.reshape(-1, 1)).sum(dim=1)

        tau_ind_next = tau[range(N), ind]
        tau_ind = tau_ind_next - dt[range(N), ind]

        # ind = torch.clip(torch.floor_divide(t, self.dt).to(
        # torch.long), min=None, max=self.N-1)

        return Discretization(
            ind=ind,
            dt=dt,
            dt_ind=dt[range(N), ind],
            tau_ind=tau_ind,
            tau_next_ind=tau_ind_next), z



r"""Datasets included with PySaRe

Datasets
--------
- ``WeibullUniformParameters`` -- A simulated datases with Weibull distributed
  event times  
"""

from .Dataset import Dataset
from . import datasets




from __future__ import annotations
import numpy as np
import torch
import warnings
from collections.abc import Iterable
import warnings
from pysare._typing import X, T, E
from typing import Optional, Union, Iterable, Callable, Literal, TypeVar, NewType, Any, List
from typing import Tuple
from torch import Tensor
from torch.nn import Module
import torch
import torch.utils
import torch.utils.data
import torch.utils.data.sampler
import numpy as np
from numpy.typing import NDArray

# from pysare.models.SurvivalModel import TorchSurvivalModel


def try_find_parameter_type(torch_model):
    # Loop through parameters until a floating point
    for parameter in torch_model.parameters():
        if torch.is_floating_point(parameter):
            return parameter.dtype, parameter.device

    # If model is a distribution no warning is given
    from pysare.models.distributions import _distributions
    if not torch_model.__class__ in _distributions:
        warnings.warn(
            "Could not determine device and dtype from self.parameters()," +
            "using the default tensor type")

    # Returing default tensor type
    return torch.empty(0).dtype, torch.empty(0).device


def convert_to_tensor(data, dtype, device):
    if torch.is_tensor(data):
        return data.to(dtype=dtype, device=device)
    else:
        return torch.tensor(data, dtype=dtype, device=device)


INDEX_DTYPE = torch.long


def validate_inputs(batch_size, max_time, X, T, E=None):

    # If E is provided it must have same shape as T
    if (not E is None) and (not E.shape == T.shape):
        raise ValueError("The shape of T and E must be equal")

    if len(X.shape) < 2:
        raise ValueError(
            "The number of dimensions of X must be larger than 1.")

    N = X.shape[0]

    squeeze = False

    if len(T.shape) > 2:
        raise ValueError(
            "The number of dimensions of T must be either 1 or 2.")

    elif len(T.shape) == 0:
        # T is scalar
        # ----------------------------------------------------------------------

        T_inds = torch.full((N,), 0, dtype=INDEX_DTYPE, device=T.device)
        X_inds = torch.arange(0, N, 1, dtype=INDEX_DTYPE, device=X.device)
        T_out_inds = T_inds.cpu()
        X_out_inds = X_inds.cpu()

        if N > 1:
            output_shape = (N, 1)
        else:
            output_shape = (1, 1)
            squeeze = True

        T = T.reshape(-1,)

    elif len(T.shape) == 1:
        # T has shape (M,)
        # ----------------------------------------------------------------------
        M = T.shape[0]
        T_inds = torch.arange(0, M, 1, dtype=INDEX_DTYPE,
                              device=T.device).repeat(N)
        X_inds = torch.arange(0, N, dtype=INDEX_DTYPE,
                              device=X.device).repeat_interleave(M)
        T_out_inds = T_inds.cpu()
        X_out_inds = X_inds.cpu()
        if N == 1:
            output_shape = (1, M)
            squeeze = True
        else:
            output_shape = (N, M)

    elif T.shape[0] == 1:
        # T has shape (1,M)
        # ----------------------------------------------------------------------
        M = T.shape[1]
        T_inds = torch.arange(0, M, 1, dtype=INDEX_DTYPE,
                              device=T.device).repeat(N)
        X_inds = torch.arange(0, N, dtype=INDEX_DTYPE,
                              device=X.device).repeat_interleave(M)
        T_out_inds = T_inds.cpu()
        X_out_inds = X_inds.cpu()
        output_shape = (N, M)

    elif T.shape[0] == X.shape[0]:
        if T.shape[1] != 1:
            raise ValueError("Either T.shape[0] or T.shape[1] must be 1.")
        # T has shape (N,1)
        # ----------------------------------------------------------------------
        # T is a column vector for witch row i is applied to row i in X
        T_inds = torch.arange(0, N, 1, dtype=INDEX_DTYPE, device=T.device)
        X_inds = torch.arange(0, N, 1, dtype=INDEX_DTYPE, device=X.device)
        T_out_inds = torch.full((N,), 0, dtype=INDEX_DTYPE, device='cpu')
        X_out_inds = X_inds.cpu()

        output_shape = (N, 1)
    else:
        raise ValueError(
            "T.shape[0] must be equal to 1 or equal to X.shape[0].")

    left_inds = T[T_inds].reshape(-1,) < 0
    right_inds = T[T_inds].reshape(-1,) > max_time
    mid_inds = ~left_inds & ~right_inds

    if batch_size is None:
        batch_size = len(X_inds)

    # return X_inds, X_out_inds, T_inds, T_out_inds
    return ([[data[inds].split(batch_size)
             for inds in (left_inds, mid_inds, right_inds)]
            for data in (X_inds, X_out_inds, T_inds, T_out_inds)],
            output_shape,
            squeeze,
            batch_size)


class NumpySurvivalModel():
    r"""NumPy interface to PyTorch Model

    Implements an interface to a PyTorch model creating a survival model that
    takes NumPy arrays (or any object that can be converted to such) as input.

    Typically created by calling the method `to_numpy` on a PyTorch Model.

    Parameters
    ----------

    torch_model : TorchSurvivalModel
        The original PyTorch model to be interfaced
    torch_max_time : float, optional
        Maximal time for which the original PyTorch model is defined. The
        By default, the support torch_model.max_time() is used.
    numpy_support : tuple(float, float), optional
        Support of the constructed NumPy model. The numpy model will map
        [0.0, torch_max_time] onto numpy_support using an affine
        transformation. By default the same support as the original PyTorch
        model is used.
    batch_size : int, optional
        Maximal batch size to be used when calling the underlying PyTorch
        Model. By default an infinite batch size, or the size of the
        DataLoader optionally specified as the data parameter.
    data : tuple(X, T, E) or Dataset or DataLoader, optional
        Data used to initialize the interface. If data should represent the
        tensors X, T, and E, and the device and dtype of these tensors
        are used to determine the correct tensor types for the torch_model.
        If data is a DataLoader its batch size will also be used in the
        interface, if not batch_size is set.
    extrapolation : {'set_zero', 'model'}, default='set_zero'
        Determines how the model is extrapolated beyond torch_max_time.
        - 'set_zero' -- Survival function is set to zero
        - 'model' -- The value given by the torch_model is used even beyond
            its maximal time. Can result in unexpected behavior and should be
            used with caution.
    """

    def __init__(self,
                 torch_model: Any,
                 torch_max_time: Optional[float] = None,
                 numpy_support: Optional[Tuple[float, float]] = None,
                 batch_size: Optional[int] = None,
                 data: Optional[Union[Tuple,
                                      pysare.data.Dataset,
                                      torch.utils.data.DataLoader]] = None,
                 extrapolation: Literal['set_zero', 'model'] = 'set_zero'):

        if data is None:

            dtype, device = try_find_parameter_type(torch_model)

            def data_transform(X, T, E):
                return (convert_to_tensor(X, dtype=dtype, device=device),
                        convert_to_tensor(T, dtype=dtype, device=device),
                        convert_to_tensor(E, dtype=torch.bool, device=device))
        else:
            if isinstance(data, Iterable):
                temp = next(iter(data))
                if isinstance(temp, Union[List,Tuple]):
                    # data is a data loader/iterates over batches
                    X, T, E = temp
                else:
                    # data is a tuple (X, T, E)
                    X, T, E = data
            else:
                # Data is Dataset
                X, T, E = data[:]

            X_device = X.device
            X_dtype = X.dtype
            T_device = T.device
            T_dtype = T.dtype
            E_device = E.device

            def data_transform(X, T, E):
                return (convert_to_tensor(X,
                                          dtype=X_dtype,
                                          device=X_device),
                        convert_to_tensor(T,
                                          dtype=T_dtype,
                                          device=T_device),
                        convert_to_tensor(E,
                                          dtype=torch.bool,
                                          device=E_device))

        self.data_transform = data_transform
        self.torch_model = torch_model

        # Batch size
        if (batch_size is None) and hasattr(data, 'batch_size'):
            self.batch_size = data.batch_size
        else:
            self.batch_size = batch_size

        # Support
        if torch_max_time is None:
            self.torch_max_time = torch_model.max_time()
        else:
            self.torch_max_time = torch_max_time
        if numpy_support is None:
            self.numpy_support = (0, self.torch_max_time)
        else:
            if not numpy_support[1] < np.inf:
                raise ValueError("Can not transform support of model with" +
                                 "infinite support.")
            self.numpy_support = tuple(numpy_support)

        self.numpy_min_time = self.numpy_support[0]
        if self.torch_max_time < np.inf:
            self.numpy_scale = self.torch_max_time/(self.numpy_support[1]
                                                    - self.numpy_support[0])
        else:
            self.numpy_scale = 1.0

        self.extrapolation = extrapolation

    def freeze(self, X: Any) -> FrozenNumpySurvivalModel:
        r"""Freezes the model around X,

        Freezes the model around a set of covariates/features represented by
        the ArrayLike parameter X"""
        with torch.no_grad():
            self.torch_model.eval()
            X, _, _ = self.data_transform(X, torch.empty(0), torch.empty(0))
            if self.batch_size is None:
                batch_size = X.shape[0]
            else:
                batch_size = self.batch_size

            forward_out = torch.concat([self.torch_model.forward(X_batch)
                                        for X_batch in X.split(batch_size, dim=0)],
                                       dim=0)

            return FrozenNumpySurvivalModel(self, forward_out)

    def to_torch(self) -> TorchSurvivalModel:
        r"""Returns the underlying PyTorch Model."""
        return self.torch_model

    def max_time(self) -> float:
        r"""Returns the maximal time after which extrapolation is used."""
        return self.numpy_support[1]

    def log_likelihood(self, X: Any, T: Any, E: Any) -> NDArray:
        r"""Returns the log-likelihood given X, T, and E.

        Parameters
        ----------

        X : (N,...) array_like
            Array with features/covariates with at least two dimensions,
            where the first dimension is the batch dimension of size N.
        T : (N, 1) or (N, 1) or (N,) array_like
            Array of times to be evaluated. The shape of T should be
            (N,1) (same first dimension of X), (1, M), or (M,), evaluated
            as follows:

            - (N, 1) Each element in the first dimension of X is evaluated with
            the corresponding element in T, creating an output with shape (M,1)
            - (1, M) -- Each element in the first dimension of X is evaluated
            with all elements in T creating an output with shape (N, M)
            - (M,) -- Same as for (1, M), with the extension that if M=1 the
            shape of the output is (M,)

        E : (N, 1) or (N, 1) or (N,) bool array_like
            Array of event types with the same size as T. Entries evaluated
            as True indicates recorded events and entries evaluated as False
            indicate censored times"""

        self.torch_model.eval()
        with torch.no_grad():

            X, T, E = self.data_transform(X, T, E)

            T = (T-self.numpy_min_time)*self.numpy_scale

            ((X_ind_batches, X_out_ind_batches,
              T_ind_batches, T_out_ind_batches),
             output_size, squeeze, batch_size) = validate_inputs(self.batch_size,
                                                                 self.torch_max_time,
                                                                 X,
                                                                 T)

            T = T.reshape(-1,)
            E = E.reshape(-1,)

            forward_out = torch.concat([self.torch_model.forward(X_batch)
                                        for X_batch in X.split(batch_size, dim=0)],
                                       dim=0)

            return self._log_likelihood(forward_out, T, E,
                                        X_ind_batches, X_out_ind_batches,
                                        T_ind_batches, T_out_ind_batches,
                                        output_size, squeeze)

    def _log_likelihood(self, forward_out, T, E,
                        X_ind_batches, X_out_ind_batches,
                        T_ind_batches, T_out_ind_batches,
                        output_size, squeeze):

        log_lik = np.full(output_size, np.NINF)

        # T<0 inds
        for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[0],
                                                          X_out_ind_batches[0],
                                                          T_ind_batches[0],
                                                          T_out_ind_batches[0]):

            inds = ~E[T_inds]
            if len(inds):
                log_lik[X_out_inds[inds], T_out_inds[inds]] = np.log(1.0)

        # T in support inds
        for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[1],
                                                          X_out_ind_batches[1],
                                                          T_ind_batches[1],
                                                          T_out_ind_batches[1]):
            log_lik[X_out_inds, T_out_inds] = self.torch_model._log_likelihood(
                forward_out[X_inds], T[T_inds], E[T_inds]).numpy()

        # T>max_time inds
        if self.extrapolation == 'model':
            for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[2],
                                                              X_out_ind_batches[2],
                                                              T_ind_batches[2],
                                                              T_out_ind_batches[2]):
                log_lik[X_out_inds, T_out_inds] = self.torch_model._log_likelihood(
                    forward_out[X_inds], T[T_inds], E[T_inds]).numpy()

        # Add shape constant to events
        log_lik[E.cpu()] += np.log(self.numpy_scale)

        if squeeze:
            return log_lik.squeeze()
        else:
            return log_lik

    def density_function(self, X: Any, T: Any) -> NDArray:
        r"""Returns the density at T given X.

        Parameters
        ----------

        X : (N,...) array_like
            Array with features/covariates with at least two dimensions,
            where the first dimension is the batch dimension of size N.
        T : (N, 1) or (N, 1) or (N,) array_like
            Array of times to be evaluated. The shape of T should be
            (N,1) (same first dimension of X), (1, M), or (M,), evaluated
            as follows:

            - (N, 1) Each element in the first dimension of X is evaluated with
            the corresponding element in T, creating an output with shape (M,1)
            - (1, M) -- Each element in the first dimension of X is evaluated
            with all elements in T creating an output with shape (N, M)
            - (M,) -- Same as for (1, M), with the extension that if M=1 the
            shape of the output is (M,)
            """
        self.torch_model.eval()
        with torch.no_grad():

            X, T, _ = self.data_transform(X, T, torch.empty(0))

            T = (T-self.numpy_min_time)*self.numpy_scale

            ((X_ind_batches, X_out_ind_batches,
              T_ind_batches, T_out_ind_batches),
             output_size, squeeze, batch_size) = validate_inputs(self.batch_size,
                                                                 self.torch_max_time,
                                                                 X,
                                                                 T)

            T = T.reshape(-1,)

            forward_out = torch.concat([self.torch_model.forward(X_batch)
                                        for X_batch in X.split(batch_size, dim=0)],
                                       dim=0)

            return self._density_function(forward_out, T,
                                          X_ind_batches, X_out_ind_batches,
                                          T_ind_batches, T_out_ind_batches,
                                          output_size, squeeze)

    def _density_function(self, forward_out, T,
                          X_ind_batches, X_out_ind_batches,
                          T_ind_batches, T_out_ind_batches,
                          output_size, squeeze):

        f = np.zeros(output_size)

        # T in support inds
        for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[1],
                                                          X_out_ind_batches[1],
                                                          T_ind_batches[1],
                                                          T_out_ind_batches[1]):
            f[X_out_inds, T_out_inds] = self.torch_model._density_function(
                forward_out[X_inds], T[T_inds]).numpy().squeeze()*self.numpy_scale

        # T>max_time inds
        if self.extrapolation == 'model':
            for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[2],
                                                              X_out_ind_batches[2],
                                                              T_ind_batches[2],
                                                              T_out_ind_batches[2]):
                f[X_out_inds, T_out_inds] = self.torch_model._density_function(
                    forward_out[X_inds], T[T_inds]).numpy().squeeze()*self.numpy_scale

        if squeeze:
            return f.squeeze()
        else:
            return f

    def survival_function(self, X, T) -> NDArray:
        r"""Returns the survival function at T given X.

        Parameters
        ----------

        X : (N,...) array_like
            Array with features/covariates with at least two dimensions,
            where the first dimension is the batch dimension of size N.
        T : (N, 1) or (N, 1) or (N,) array_like
            Array of times to be evaluated. The shape of T should be
            (N,1) (same first dimension of X), (1, M), or (M,), evaluated
            as follows:

            - (N, 1) Each element in the first dimension of X is evaluated with
            the corresponding element in T, creating an output with shape (M,1)
            - (1, M) -- Each element in the first dimension of X is evaluated
            with all elements in T creating an output with shape (N, M)
            - (M,) -- Same as for (1, M), with the extension that if M=1 the
            shape of the output is (M,)
            """
        self.torch_model.eval()
        with torch.no_grad():
            X, T, _ = self.data_transform(X, T, torch.empty(0))

            T = (T-self.numpy_min_time)*self.numpy_scale

            ((X_ind_batches, X_out_ind_batches,
              T_ind_batches, T_out_ind_batches),
             output_size, squeeze, batch_size) = validate_inputs(self.batch_size,
                                                                 self.torch_max_time,
                                                                 X,
                                                                 T)

            T = T.reshape(-1,)

            forward_out = torch.concat([self.torch_model.forward(X_batch)
                                        for X_batch in X.split(batch_size, dim=0)],
                                       dim=0)
            return self._survival_function(forward_out, T,
                                           X_ind_batches, X_out_ind_batches,
                                           T_ind_batches, T_out_ind_batches,
                                           output_size, squeeze)

    def _survival_function(self, forward_out, T,
                           X_ind_batches, X_out_ind_batches,
                           T_ind_batches, T_out_ind_batches,
                           output_size, squeeze):
        S = np.zeros(output_size)
        # T<0 inds
        for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[0],
                                                          X_out_ind_batches[0],
                                                          T_ind_batches[0],
                                                          T_out_ind_batches[0]):
            S[X_out_inds, T_out_inds] = 1.0

        # T in support inds
        for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[1],
                                                          X_out_ind_batches[1],
                                                          T_ind_batches[1],
                                                          T_out_ind_batches[1]):
            S[X_out_inds, T_out_inds] = self.torch_model._survival_function(
                forward_out[X_inds], T[T_inds]).numpy().squeeze()

        # T>max_time inds
        if self.extrapolation == 'model':
            for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[2],
                                                              X_out_ind_batches[2],
                                                              T_ind_batches[2],
                                                              T_out_ind_batches[2]):
                S[X_out_inds, T_out_inds] = self.torch_model._survival_function(
                    forward_out[X_inds], T[T_inds]).numpy().squeeze()

        if squeeze:
            return S.squeeze()
        else:
            return S

    def hazard_function(self, X: Any, T: Any) -> NDArray:
        r"""Returns the hazard at T given X.

        Parameters
        ----------

        X : (N,...) array_like
            Array with features/covariates with at least two dimensions,
            where the first dimension is the batch dimension of size N.
        T : (N, 1) or (N, 1) or (N,) array_like
            Array of times to be evaluated. The shape of T should be
            (N,1) (same first dimension of X), (1, M), or (M,), evaluated
            as follows:

            - (N, 1) Each element in the first dimension of X is evaluated with
            the corresponding element in T, creating an output with shape (M,1)
            - (1, M) -- Each element in the first dimension of X is evaluated
            with all elements in T creating an output with shape (N, M)
            - (M,) -- Same as for (1, M), with the extension that if M=1 the
            shape of the output is (M,)
            """

        self.torch_model.eval()
        with torch.no_grad():

            X, T, _ = self.data_transform(X, T, torch.empty(0))

            T = (T-self.numpy_min_time)*self.numpy_scale

            ((X_ind_batches, X_out_ind_batches,
              T_ind_batches, T_out_ind_batches),
             output_size, squeeze, batch_size) = validate_inputs(self.batch_size,
                                                                 self.torch_max_time,
                                                                 X,
                                                                 T)

            T = T.reshape(-1,)

            forward_out = torch.concat([self.torch_model.forward(X_batch)
                                        for X_batch in X.split(batch_size, dim=0)],
                                       dim=0)

            return self._hazard_function(forward_out, T,
                                         X_ind_batches, X_out_ind_batches,
                                         T_ind_batches, T_out_ind_batches,
                                         output_size, squeeze)

    def _hazard_function(self, forward_out, T,
                         X_ind_batches, X_out_ind_batches,
                         T_ind_batches, T_out_ind_batches,
                         output_size, squeeze):

        h = np.zeros(output_size)

        # T in support inds
        for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[1],
                                                          X_out_ind_batches[1],
                                                          T_ind_batches[1],
                                                          T_out_ind_batches[1]):
            h[X_out_inds, T_out_inds] = self.torch_model._hazard_function(
                forward_out[X_inds], T[T_inds]).numpy().squeeze()*self.numpy_scale

        # T>max_time inds
        if self.extrapolation == 'model':
            for X_inds, X_out_inds, T_inds, T_out_inds in zip(X_ind_batches[2],
                                                              X_out_ind_batches[2],
                                                              T_ind_batches[2],
                                                              T_out_ind_batches[2]):
                h[X_out_inds, T_out_inds] = self.torch_model._hazard_function(
                    forward_out[X_inds], T[T_inds]).numpy().squeeze()*self.numpy_scale

        if squeeze:
            return h.squeeze()
        else:
            return h

    def sample(self, X: Any, num_samples: int = 1,
               generator: Optional[torch.Generator] = None,
               **kwargs):
        r"""Generate samples from the distribution

        Parameters
        ----------

        X : (N,...) array_like
            Array with features/covariates with at least two dimensions,
            where the first dimension is the batch dimension of size N.

        num_samples : int, default=1
            Number of samples per individual. 

        generator : torch.Generator, optional
            Pseudorandom number generator used for the sampling. If None,
            the default generator is used.
        """

        with torch.no_grad():
            X, T_dummy, _ = self.data_transform(
                X, torch.empty(0), torch.empty(0))

            if len(X.shape) < 2:
                raise ValueError(
                    "The number of dimensions of X must be larger than 1.")

            N = X.shape[0]

            if self.batch_size is None:
                batch_size = N
            else:
                batch_size = self.batch_size

            T = np.empty((N, num_samples))

            for inds in torch.arange(0, N, 1,
                                     dtype=INDEX_DTYPE,
                                     device=T_dummy.device).split(batch_size):

                T[inds, :] = self.torch_model.sample(
                    X[inds], num_samples, generator,**kwargs)

            return T / self.numpy_scale + self.numpy_min_time


class FrozenNumpySurvivalModel():
    r"""Freezes a NumPy model around a set of covariates/features

    Parameters
    ----------

    numpy_model : NumpyModel
        The NumPy model to be frozen
    forward_out : (M,...) Tensor
        The output from the forward function for the covariates/features
        to freeze the model around.
    """

    def __init__(self,
                 numpy_model: NumpySurvivalModel,
                 forward_out: Tensor):
        self.numpy_model = numpy_model
        self.forward_out = forward_out

    def __len__(self):
        return self.forward_out.shape[0]

    def __getitem__(self, inds) -> FrozenNumpySurvivalModel:
        return FrozenNumpySurvivalModel(self.numpy_model,
                                        self.forward_out[inds].reshape(-1, *self.forward_out.shape[1:]))

    def max_time(self) -> float:
        r"""Returns the maximal time after which extrapolation is used."""
        return self.numpy_model.max_time()

    def to_torch(self) -> TorchSurvivalModel:
        r"""Unfreezes the model and returns the underlying PyTorch model"""
        return self.numpy_model.to_torch()

    def unfreeze(self,) -> NumpySurvivalModel:
        r"""Unfreezes the model and returns the underlying NumPy model"""
        return self.numpy_model

    def log_likelihood(self, T: Any, E: Any) -> NDArray:
        r"""Returns the survival function at T.

        Parameters
        ----------

        T : (N, 1) or (N, 1) or (N,) array_like
            Array of times to be evaluated. The shape of T should be
            (N,1) (same first dimension of X), (1, M), or (M,), evaluated
            as follows:

            - (N, 1) Each element in the first dimension of X is evaluated with
            the corresponding element in T, creating an output with shape (M,1)
            - (1, M) -- Each element in the first dimension of X is evaluated
            with all elements in T creating an output with shape (N, M)
            - (M,) -- Same as for (1, M), with the extension that if M=1 the
            shape of the output is (M,)

        E : (N, 1) or (N, 1) or (N,) bool array_like
            Array of event types with the same size as T. Entries evaluated
            as True indicates recorded events and entries evaluated as False
            indicate censored times
            """

        with torch.no_grad():
            _, T, E = self.numpy_model.data_transform(
                torch.empty(0), T, E)

            T = (T-self.numpy_model.numpy_min_time) * \
                self.numpy_model.numpy_scale

            ((X_ind_batches, X_out_ind_batches,
              T_ind_batches, T_out_ind_batches),
             output_size, squeeze, _) = validate_inputs(self.numpy_model.batch_size,
                                                        self.numpy_model.torch_max_time,
                                                        self.forward_out,
                                                        T)

            T = T.reshape(-1,)
            E = E.reshape(-1,)

            return self.numpy_model._log_likelihood(self.forward_out, T, E,
                                                    X_ind_batches, X_out_ind_batches,
                                                    T_ind_batches, T_out_ind_batches,
                                                    output_size, squeeze)

    def density_function(self, T) -> NDArray:
        r"""Returns the density function at T.

        Parameters
        ----------

        T : (N, 1) or (N, 1) or (N,) array_like
            Array of times to be evaluated. The shape of T should be
            (N,1) (same first dimension of X), (1, M), or (M,), evaluated
            as follows:

            - (N, 1) Each element in the first dimension of X is evaluated with
            the corresponding element in T, creating an output with shape (M,1)
            - (1, M) -- Each element in the first dimension of X is evaluated
            with all elements in T creating an output with shape (N, M)
            - (M,) -- Same as for (1, M), with the extension that if M=1 the
            shape of the output is (M,)
            """

        with torch.no_grad():
            _, T, _ = self.numpy_model.data_transform(
                torch.empty(0), T, torch.empty(0))

            T = (T-self.numpy_model.numpy_min_time) * \
                self.numpy_model.numpy_scale

            ((X_ind_batches, X_out_ind_batches,
              T_ind_batches, T_out_ind_batches),
             output_size, squeeze, _) = validate_inputs(self.numpy_model.batch_size,
                                                        self.numpy_model.torch_max_time,
                                                        self.forward_out,
                                                        T)

            T = T.reshape(-1,)

            return self.numpy_model._density_function(self.forward_out, T,
                                                      X_ind_batches, X_out_ind_batches,
                                                      T_ind_batches, T_out_ind_batches,
                                                      output_size, squeeze)

    def survival_function(self, T) -> NDArray:
        r"""Returns the survival function at T.

        Parameters
        ----------

        T : (N, 1) or (N, 1) or (N,) array_like
            Array of times to be evaluated. The shape of T should be
            (N,1) (same first dimension of X), (1, M), or (M,), evaluated
            as follows:

            - (N, 1) Each element in the first dimension of X is evaluated with
            the corresponding element in T, creating an output with shape (M,1)
            - (1, M) -- Each element in the first dimension of X is evaluated
            with all elements in T creating an output with shape (N, M)
            - (M,) -- Same as for (1, M), with the extension that if M=1 the
            shape of the output is (M,)
            """

        with torch.no_grad():
            _, T, _ = self.numpy_model.data_transform(
                torch.empty(0), T, torch.empty(0))

            T = (T-self.numpy_model.numpy_min_time) * \
                self.numpy_model.numpy_scale

            ((X_ind_batches, X_out_ind_batches,
              T_ind_batches, T_out_ind_batches),
             output_size, squeeze, _) = validate_inputs(self.numpy_model.batch_size,
                                                        self.numpy_model.torch_max_time,
                                                        self.forward_out,
                                                        T)

            T = T.reshape(-1,)

            return self.numpy_model._survival_function(self.forward_out, T,
                                                       X_ind_batches, X_out_ind_batches,
                                                       T_ind_batches, T_out_ind_batches,
                                                       output_size, squeeze)

    def hazard_function(self, T: Any) -> NDArray:
        r"""Returns the hazard function at T.

        Parameters
        ----------

        T : (N, 1) or (N, 1) or (N,) array_like
            Array of times to be evaluated. The shape of T should be
            (N,1) (same first dimension of X), (1, M), or (M,), evaluated
            as follows:

            - (N, 1) Each element in the first dimension of X is evaluated with
            the corresponding element in T, creating an output with shape (M,1)
            - (1, M) -- Each element in the first dimension of X is evaluated
            with all elements in T creating an output with shape (N, M)
            - (M,) -- Same as for (1, M), with the extension that if M=1 the
            shape of the output is (M,)
            """

        with torch.no_grad():
            _, T, _ = self.numpy_model.data_transform(
                torch.empty(0), T, torch.empty(0))

            T = (T-self.numpy_model.numpy_min_time) * \
                self.numpy_model.numpy_scale

            ((X_ind_batches, X_out_ind_batches,
              T_ind_batches, T_out_ind_batches),
             output_size, squeeze, _) = validate_inputs(self.numpy_model.batch_size,
                                                        self.numpy_model.torch_max_time,
                                                        self.forward_out,
                                                        T)

            T = T.reshape(-1,)

            return self.numpy_model._hazard_function(self.forward_out, T,
                                                     X_ind_batches, X_out_ind_batches,
                                                     T_ind_batches, T_out_ind_batches,
                                                     output_size, squeeze)

    def sample(self, num_samples: int = 1,
               generator: Optional[torch.Generator] = None,
               **kwargs):
        r"""Generate samples from the distribution

        Parameters
        ----------

        num_samples : int, default=1
            Number of samples per individual. 

        generator : torch.Generator, optional
            Pseudorandom number generator used for the sampling. If None,
            the default generator is used.
        """

        with torch.no_grad():

            N = self.forward_out.shape[0]

            if self.numpy_model.batch_size is None:
                batch_size = N
            else:
                batch_size = self.numpy_model.batch_size

            T = np.empty((N, num_samples))

            for inds in torch.arange(0, N, 1,
                                     dtype=INDEX_DTYPE,
                                     device=self.forward_out.device).split(batch_size):

                T[inds, :] = self.numpy_model.torch_model._sample(self.forward_out[inds],
                                                                  num_samples,
                                                                  generator,**kwargs)

            return T / self.numpy_model.numpy_scale + self.numpy_model.numpy_min_time



r"""PySaRe is a package for Data-Driven Survival Analysis and Reliability Engineering based on PyTorch.

Submodules
----------
- ``models`` -- Implementations of survival models
- ``data`` -- Tools for handling data and some included datasets
- ``training`` -- Tools for training models 
- ``evaluate`` -- Evaluation metrics
"""

from . import data
from . import models
from . import training
from . import reliability
from . import evaluation
from . import age_replacement


# %%
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from dataclasses import dataclass
from collections import namedtuple
import math
import scipy

# %%
# t = policy.t
# r = policy.r
# r[r==0] = .6


def plot_policy(h, r, ax, color=None, tol=2, resolution=np.inf, plot_opt='-', inf_level=None):

    r = r.copy()

    if any(r > 0):
        r_min = r[r > 0].min()
    else:
        r_min = 0

    if inf_level is None:
        r_inf = r.max() + (r.max()-r_min)*.2
    else:
        r_inf = inf_level
    r[r == 0] = r_inf

    # downsample if enough points
    if len(h) > 2*resolution:
        t_min = min(h[r != np.inf])
        t_max = max(h[r != np.inf])

        samples_per_time_unit = resolution/(t_max-t_min)
    else:
        samples_per_time_unit = False

    def find_discontinuities(t, r, tol=2):
        dr_tol = t[1]-t[0]

        dr = abs(np.diff(r, prepend=r[0])).clip(dr_tol)
        dr_windowed = np.min((dr[:-2], dr[1:-1], dr[2:]), axis=0)
        dr_windowed = np.insert(dr_windowed, (0, len(
            dr_windowed)), (dr_windowed[0], dr_windowed[-1]), axis=0)

        return np.where(dr/dr_windowed > tol)[0]

    def find_pieces(t, r, tol=2):
        disc = find_discontinuities(t, r, tol)

        return np.insert(disc, (0, len(disc)), (0, len(t)-1), axis=0)

    pieces = find_pieces(h, r, tol)

    if color is None:
        line, = ax.plot(np.nan, np.nan)
        color = line.get_color()

    for n in range(len(pieces)-1):
        start = pieces[n]
        stop = pieces[n+1]

        if samples_per_time_unit:
            inds = np.round(np.linspace(
                start, stop-1, int(np.ceil((h[stop]-h[start])*samples_per_time_unit)+1))).astype(int)
            ax.plot(h[inds], r[inds], plot_opt, color=color)
        else:
            ax.plot(h[start:stop], r[start:stop], plot_opt, color=color)

    # Wierd code to get the correct yticklabels
    labels = [item.get_text() for item in ax.get_yticklabels()
              if item.get_position()[1] < r_inf]
    ticks = [item.get_position()[1] for item in ax.get_yticklabels()
             if item.get_position()[1] < r_inf]

    ticks += [r_inf]
    labels += [r"$\infty$"]
    # ax.set_yticks([tick for tick in ax.get_yticks() if tick < r_inf] + [r_inf])
    ax.set_yticks(ticks)

    print(labels)
    print(ticks)
    # labels = [item.get_text() for item in ax.get_yticklabels()]
    # labels += ['$\infty$']
    ax.set_yticklabels(labels)
    # plot_policy(t,r,plt,tol=1.1)

# def plot_policy(t, r, ax, color=None, tol=2):

#     r = r.copy()

#     def find_discontinuities(t, r, tol = 2):
#         dr_tol = t[1]-t[0]


#         dr = abs(np.diff(r,prepend=r[0])).clip(dr_tol)
#         dr_windowed = np.min((dr[:-2],dr[1:-1],dr[2:]),axis=0)
#         dr_windowed = np.insert(dr_windowed,(0, len(dr_windowed)),(dr_windowed[0], dr_windowed[-1]),axis = 0)

#         return np.where(dr/dr_windowed>tol)[0]

#     def find_pieces(t,r,tol=2):
#         disc = find_discontinuities(t,r,tol)

#         return np.insert(disc,(0, len(disc)),(0, len(t)-1),axis = 0)

#     pieces = find_pieces(t,r)

#     if color is None:
#         line, = ax.plot(np.nan,np.nan)
#         color = line.get_color()


#     if any(r>0):
#         r_min = r[r>0].min()
#     else:
#         r_min = 0

#     r_inf = r.max() + (r.max()-r_min)*.2
#     r[r==0] = r_inf


#     for n in range(len(pieces)-1):
#         start = pieces[n]
#         stop = pieces[n+1]

#         ax.plot(t[start:stop],r[start:stop],color=color)


#     # Wierd code to get the correct yticklabels
#     labels = [item.get_text() for item in ax.get_yticklabels() if item.get_position()[1]<r_inf]
#     ticks = [item.get_position()[1] for item in ax.get_yticklabels() if item.get_position()[1]<r_inf]

#     ticks += [r_inf]
#     labels +=  ['$\infty$']
#     # ax.set_yticks([tick for tick in ax.get_yticks() if tick < r_inf] + [r_inf])
#     ax.set_yticks(ticks)

#     print(labels)
#     print(ticks)
#     # labels = [item.get_text() for item in ax.get_yticklabels()]
#     # labels += ['$\infty$']
#     ax.set_yticklabels(labels)
#     # plot_policy(t,r,plt,tol=1.1)

def renewal_function(S, t_m, N):
    t = np.linspace(0, t_m, N)
    m = np.zeros_like(t)

    F = 1 - S(t)
    dF = np.diff(F)

    for n in range(2, N):
        m[n] = F[n] + np.sum(m[n - 1:: -1] * dF[:n])

    return pd.DataFrame({"t": t, "m": m}).set_index("t")


def calculate_random_horizon_age_replacement(S_f, S_h, c_f, c_p, t_m, N):
    def S_lim(t):
        Sm = S_h(t_m)
        return (S_h(t) - Sm) / (1 - Sm)

    def Sht(t, t0):
        if t0 >= t_m:
            return t * 0
        else:
            return S_lim(t + t0) / S_lim(np.array([t0]))

        S = S_h(t + t0) / S_h(t0)
        S[t0 >= tm] = 0
        return S

    t = np.linspace(0, t_m, N)

    C = np.zeros_like(t)
    r = np.zeros_like(C)

    # C[-1] = 1.0 # HERE----------------------------------------------------

    for n in range(N - 2, -1, -1):
        km = N - n - 1

        Shn = Sht(t[: km], t[n])
        Sfn = S_f(t[1: km + 1])

        dSf = S_f(t[:km]) - S_f(t[1: km + 1])

        Cnk = C[n + 1:]

        Cnf = np.cumsum((Cnk + c_f) * Shn * dSf)
        Cnp = (c_p + Cnk) * Shn * Sfn
        Cn = Cnf + Cnp

        ind = np.argmin(Cn)
        C[n] = Cn[ind]
        r[n] = ind + 1

        if Cnf[-1] <= C[n]:
            C[n] = Cnf[-1]
            r[n] = km

        if r[n] == km:
            r[n] = 0
    return pd.DataFrame({"t": t, "C": C, "r": r, "S_lim": S_lim(t)}).set_index("t")


def random_horizon_bounds(S_f, S_h, c_f, c_p, t_m, N):
    def Sht(t, t0):
        if t0 >= t_m:
            return t * 0
        else:
            return S_h(t+t0) / S_h(np.array([t0]))

    t = np.linspace(0, t_m, N)

    m = renewal_function(S_f, t_m, N)["m"].values[:-1]

    c_ub = np.zeros(N)
    for n in range(N):
        dF = np.diff(-Sht(t, t[n]))
        c_ub[n] = np.sum(dF * m)

    return pd.DataFrame({"ub": c_ub * c_f, "lb": c_ub * c_p, "t": t}).set_index("t")


def finit_support_bound(S_f, S_h, c_f, c_p, t_0, t_m, N_0):

    def Sht(t, t0):
        if t0 >= t_m:
            return t * 0
        else:
            return S_h(t+t0) / S_h(np.array([t0]))

    dt = t_0/(N_0-1)
    t = np.arange(0, t_m+dt/2, dt)

    m = renewal_function(S_f, t[-1], len(t))['m'].values

    B = np.zeros(N_0)

    for n in range(N_0):

        dF = np.diff(-S_h(t[N_0:])/S_h(t[n]))
        B[n] = np.sum(dF*m[N_0-n:N_0-n+len(dF)])

    return pd.DataFrame({"B": c_f*B, "L": c_p*B, "t": t[:N_0]}).set_index("t")


def finit_support_ractive(S_f, S_h, c_f, t_0, t_m, N_0):

    def Sht(t, t0):
        if t0 >= t_m:
            return t * 0
        else:
            return S_h(t+t0) / S_h(np.array([t0]))

    dt = t_0/(N_0-1)
    t = np.arange(0, t_m+dt/2, dt)

    m = renewal_function(S_f, t[-1], len(t))['m'].values

    B = np.zeros(N_0)

    for n in range(N_0):

        dF = np.diff(-S_h(t[n:N_0])/S_h(t[n]))
        B[n] = c_f*np.sum(dF*m[:len(dF)])

    return pd.DataFrame({"B": B, "t": t[:N_0]}).set_index("t")


Problem = namedtuple("Problem", "S_h S_f c_p c_f")


def find_nearest(array, value):
    idx = np.searchsorted(array, value, side="left")
    if idx > 0 and (idx == len(array) or math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
        return idx-1
    else:
        return idx


def planned_replacements(t, r):

    N = np.zeros_like(t)

    if r[-1] > 0:
        N[-1] = 1

    for n in reversed(range(len(t)-1)):

        if r[n] > 0:
            N[n] = 1 + N[find_nearest(t, t[n]+r[n])]
        else:
            N[n] = 0

    return N


class TimeBasedSequentialPolicy:
    def __init__(self, t, r, C, ub, lb, S_h, S_f) -> None:
        self.t = t
        self.r = r
        self.C = C
        self.ub = ub
        self.lb = lb
        self.S_h = S_h
        self.S_f = S_f

        self.num_planed_replacements = planned_replacements(t, r)

    def replacement_age(self, t):
        if t <= self.t[-1]:
            ind = find_nearest(self.t, t)
            if self.r[ind] > 0:
                return self.r[ind]

        return np.inf

    def plot_policy(self, ax, color=None, tol=2, resolution=np.inf, plot_opt='-', inf_level=None):
        return plot_policy(self.t, self.r, ax, color, tol, resolution, plot_opt, inf_level)


def sequential_random_horizon(failure_distribution, horizon_distribution, c_p, c_f, finite_support_approximation, num_disc_points, min_bound_approx_time):

    policy = calculate_random_horizon_age_replacement(
        failure_distribution.survival_probability, horizon_distribution.survival_probability,  c_f, c_p, finite_support_approximation, num_disc_points)

    bounds = finit_support_bound(failure_distribution.survival_probability, horizon_distribution.survival_probability,
                                 c_f, c_p, finite_support_approximation, finite_support_approximation+min_bound_approx_time, num_disc_points)

    t = policy.index.values
    S0 = horizon_distribution.survival_probability(
        finite_support_approximation)
    St = horizon_distribution.survival_probability(t)

    return TimeBasedSequentialPolicy(t,
                                     policy['r'].values*(t[1]-t[0]),
                                     policy['C'].values,
                                     policy['C'].values*(1-S0/St)+bounds['B'],
                                     np.max((policy['C'].values*(1-S0/St) +
                                             bounds['L'], policy['C']), axis=0),
                                     horizon_distribution.survival_probability(t), failure_distribution.survival_probability(t))


def simulate_single(Th, Tf, c_p, c_f, policy):
    th = Th()

    tc = 0
    c = 0
    while True:
        tf = Tf()
        rn = policy.replacement_age(tc)

        if tf < rn:
            tc += tf
            if tc >= th:
                break
            else:
                c += c_f

        else:
            tc += rn
            if tc >= th:
                break
            else:
                c += c_p

    return c


def simulate(horizon_distribution, failure_distribution, c_p, c_f, policy, N):
    C = np.zeros(N)

    for n in range(N):
        C[n] = simulate_single(horizon_distribution.simulate,
                               failure_distribution.simulate, c_p, c_f, policy)

    return C

# %%



import torch
from pysare.models.SurvivalModel import TorchSurvivalModel

class MixtureSurvivalModel(TorchSurvivalModel):
    def __init__(self, distributions, mixer) -> None:
        super().__init__()
        self.mixer = mixer
        self.distributions = distributions

    def max_time(self):
        return torch.inf

    def _survival_function(self, forward_out, T):
        ind = forward_out.shape[1]//(self.distributions.num_parameters+1)*self.distributions.num_parameters
        return self.mixer.survival_probability(T, self.distributions, forward_out[:, :ind], forward_out[:, ind:])
    
    def _log_survival_function(self, forward_out, T):
        ind = forward_out.shape[1]//(self.distributions.num_parameters+1)*self.distributions.num_parameters
        return self.mixer.log_survival_probability(T, self.distributions, forward_out[:, :ind], forward_out[:, ind:])

    def _density_function(self, forward_out, T):
        ind = forward_out.shape[1]//(self.distributions.num_parameters+1)*self.distributions.num_parameters
        return self.mixer.lifetime_density(T, self.distributions, forward_out[:, :ind], forward_out[:, ind:])

    def _log_density_function(self, forward_out, T):
        ind = forward_out.shape[1]//(self.distributions.num_parameters+1)*self.distributions.num_parameters
        return self.mixer.log_lifetime_density(T, self.distributions, forward_out[:, :ind], forward_out[:, ind:])
    
    def forward(self, X):
        return X




"""Evaluation metrics for PySaRe models.


- :func:`concordance_index` --  Concordance index (C-index) 
- ``brier_score`` -- (Integrated) Brier score 
- ``binomial_log_likelihood`` -- (Integrated) binomial log-likelihood score 
- ``negative_log_likelihood`` -- Mean negative log-likelihood
"""

from .concordance_index import concordance_index
from .brier_score import brier_score
from .binomial_log_likelihood import binomial_log_likelihood
from .negative_log_likelihood import negative_log_likelihood
from .QQ_plot import QQ_plot


import numpy as np
import matplotlib.pyplot as plt
import math


def plot_policy(h, r, ax, color=None, tol=2, resolution=np.inf, plot_opt='-', inf_level=None):

    r = r.copy()


    r[r-h>=0]=0

    if any(r > 0):
        r_min = r[r > 0].min()
    else:
        r_min = 0

    if inf_level is None:
        r_inf = r.max() + (r.max()-r_min)*.2
    else:
        r_inf = inf_level
    r[r == 0] = r_inf

    # downsample if enough points
    if len(h) > 2*resolution:
        t_min = min(h[r != np.inf])
        t_max = max(h[r != np.inf])

        samples_per_time_unit = resolution/(t_max-t_min)
    else:
        samples_per_time_unit = False

    def find_discontinuities(t, r, tol=2):
        dr_tol = t[1]-t[0]

        dr = abs(np.diff(r, prepend=r[0])).clip(dr_tol)
        dr_windowed = np.min((dr[:-2], dr[1:-1], dr[2:]), axis=0)
        dr_windowed = np.insert(dr_windowed, (0, len(
            dr_windowed)), (dr_windowed[0], dr_windowed[-1]), axis=0)

        return np.where(dr/dr_windowed > tol)[0]

    def find_pieces(t, r, tol=2):
        disc = find_discontinuities(t, r, tol)

        return np.insert(disc, (0, len(disc)), (0, len(t)-1), axis=0)

    pieces = find_pieces(h, r, tol)

    if color is None:
        line, = ax.plot(np.nan, np.nan)
        color = line.get_color()

    for n in range(len(pieces)-1):
        start = pieces[n]
        stop = pieces[n+1]

        if samples_per_time_unit:
            inds = np.round(np.linspace(
                start, stop-1, int(np.ceil((h[stop]-h[start])*samples_per_time_unit)+1))).astype(int)
            ax.plot(h[inds], r[inds], plot_opt, color=color)
        else:
            ax.plot(h[start:stop], r[start:stop], plot_opt, color=color)

    # Wierd code to get the correct yticklabels
    labels = [item.get_text() for item in ax.get_yticklabels()
              if item.get_position()[1] < r_inf]
    ticks = [item.get_position()[1] for item in ax.get_yticklabels()
             if item.get_position()[1] < r_inf]

    ticks += [r_inf]
    labels += [r"$\infty$"]
    ax.set_yticks(ticks)
    ax.set_yticklabels(labels)


def find_nearest(array, value):
    idx = np.searchsorted(array, value, side="left")
    if idx > 0 and (idx == len(array) or math.fabs(value - array[idx-1]) < math.fabs(value - array[idx])):
        return idx-1
    else:
        return idx


class HorizonBasedSequentialPolicy:
    def __init__(self, h, r, C, S) -> None:
        self.h = h
        self.r = r
        self.expected_cost = C
        self.survival_function = S

    def replacement_age(self, h):
        if h <= self.h[-1]:
            ind = find_nearest(self.h, h)
            if self.r[ind] > 0:
                return self.r[ind]

        return np.inf

    def plot_policy(self, ax, color=None, tol=2, resolution=np.inf, plot_opt='-', inf_level=None):
        return plot_policy(self.h, self.r, ax, color, tol, resolution, plot_opt, inf_level)

    def plot(self,ax=None, resolution=np.inf):
      
        # Every nth index so that in total *resolution* are plotted
        inds = self.h<0
        inds[::int(max(1,len(self.h)//resolution))] = True
        inds[-1] = True
      
        if ax is None:
            fig, ax = plt.subplots(3, 1, num=1, clear=True)
        else:
            fig=ax[0].figure
        ax[0].plot(self.h[inds],self.survival_function[inds])
        ax[0].set_xlabel('Time')
        ax[0].set_ylabel('Surv. func.')
        ax[1].plot(self.h[inds], self.expected_cost[inds])
        ax[1].set_ylabel('Cost')

        self.plot_policy(ax[2], resolution=resolution)
        fig.tight_layout()
        ax[2].set_ylabel('Policy')
        ax[2].set_xlabel('Horizon')
        return ax


def _estimate_density(S, f_in, dt):
    f = np.zeros_like(S)

    if f_in is None:
        f[0] = (S[0]-S[1])/dt
    else:
        f[0] = f_in

    for n in range(1, len(S)):
        f[n] = 2*(S[n-1]-S[n])/dt - f[n-1]

    return f


def _sequential_finite_horizon(S, h, c_p, c_f, f=None,):

    dt = h[1]-h[0]
    C = np.zeros_like(S)
    r = np.zeros_like(S)

    # If f not provided
    if not isinstance(f, np.ndarray):
        f = _estimate_density(S, f, dt)

    # Main loop
    for n in range(1, len(S)):
        Ck = np.zeros((n,))

        JC = np.flip(C[:n+1])

        l = (c_f + JC)*f[:n+1]
        L = np.cumsum(l[1:]+l[:-1])*dt/2

        Ck = (c_p+JC[1:])*S[1:n+1] + L
        C_inf = L[-1]

        min_ind = Ck.argmin()

        if Ck[min_ind] < C_inf:
            r[n] = (min_ind+1)*dt
            C[n] = Ck[min_ind]
        else:
            r[n] = (n+1)*dt
            C[n] = C_inf

    return HorizonBasedSequentialPolicy(h, r, C, S)


def sequential_finite_horizon(failure_distribution, c_p, c_f, horizon, num_disc_points):

    if hasattr(failure_distribution, 'to_numpy'):
        failure_distribution = failure_distribution.to_numpy()

    h = np.linspace(0.0, horizon, num_disc_points)

    S = np.zeros_like(h)
    S[h <= failure_distribution.max_time()] = failure_distribution.survival_function(
        h[h <= failure_distribution.max_time()])

    return _sequential_finite_horizon(S, h, c_p, c_f)

    


def _adaptive_age_replacement(Sc, dt, c_p, c_f, S=None, C=None, fc=None, f=None, interpolate=True):

    if S is not None:
        r, C = age_replacement(S, dt, c_p, c_f, f)

    C = C[:len(Sc)]

    if not isinstance(fc, np.ndarray):
        fc = estimate_density(Sc, fc, dt)

    n = len(Sc)-1

    Ck = np.zeros((n,))

    JC = np.flip(C)

    l = (c_f + JC)*fc
    L = np.cumsum(l[1:]+l[:-1])*dt/2

    Ck = (c_p+JC[1:])*Sc[1:] + L
    C_inf = L[-1]

    min_ind = Ck.argmin()

    if Ck[min_ind] < C_inf:
        rc = (min_ind+1)*dt
        Cc = Ck[min_ind]

        if interpolate:
            if min_ind > 0 and min_ind < n-1:
                x = [(min_ind)*dt, (min_ind+1)*dt, (min_ind+2)*dt]
                rc, Cc = quadratic_solution(x, Ck[min_ind-1:min_ind+2])

    else:
        rc = (n+1)*dt
        Cc = C_inf

    return rc, Cc



from torch.nn import Module
import torch


class SoftmaxMixer(Module):
    def __init__(self) -> None:
        super().__init__()

    def _calc_weights(self, log_prop_weights):
        weights = torch.exp(log_prop_weights)
        return weights/weights.sum(dim=1).reshape(-1, 1)

    def _calc_prop_weights(self, log_prop_weights):
        prop_weights = torch.exp(
            log_prop_weights - torch.max(log_prop_weights, dim=1, keepdim=True)[0])
        return prop_weights

    def survival_probability(self, T, distributions, parameters, log_prop_weights):
        weights = self._calc_weights(log_prop_weights)

        S = distributions.survival_probability(parameters, T)
        return (S*weights).sum(dim=1)

    def log_survival_probability(self, T, distributions, parameters, log_prop_weights):
        prop_weights = self._calc_prop_weights(log_prop_weights)

        S = distributions.survival_probability(parameters, T)
        return torch.log((S*prop_weights).sum(dim=1)) - torch.log(prop_weights.sum(dim=1))

    def lifetime_density(self, T, distributions, parameters, log_prop_weights):
        weights = self._calc_weights(log_prop_weights)

        f = distributions.lifetime_density(parameters, T)
        return (f*weights).sum(dim=1)

    def log_lifetime_density(self, T, distributions, parameters, log_prop_weights):
        prop_weights = self._calc_prop_weights(log_prop_weights)

        f = distributions.lifetime_density(parameters, T)
        return torch.log((f*prop_weights).sum(dim=1)) - torch.log(prop_weights.sum(dim=1))


class SoftmaxMixerWithReflection(SoftmaxMixer):
    def __init__(self) -> None:
        super().__init__()

    def survival_probability(self, T, distributions, parameters, log_prop_weights):
        weights = self._calc_weights(log_prop_weights)

        S = distributions.survival_probability(
            parameters, T) + (1-distributions.survival_probability(parameters, -T))

        return (S*weights).sum(dim=1)

    def log_survival_probability(self, T, distributions, parameters, log_prop_weights):
        prop_weights = self._calc_prop_weights(log_prop_weights)

        S = distributions.survival_probability(
            parameters, T) + (1-distributions.survival_probability(parameters, -T))

        return torch.log((S*prop_weights).sum(dim=1)) - torch.log(prop_weights.sum(dim=1))

    def lifetime_density(self, T, distributions, parameters, log_prop_weights):
        weights = self._calc_weights(log_prop_weights)

        f = distributions.lifetime_density(
            parameters, T)+distributions.lifetime_density(parameters, -T)
        return (f*weights).sum(dim=1)

    def log_lifetime_density(self, T, distributions, parameters, log_prop_weights):
        prop_weights = self._calc_prop_weights(log_prop_weights)

        f = distributions.lifetime_density(
            parameters, T)+distributions.lifetime_density(parameters, -T)
        return torch.log((f*prop_weights).sum(dim=1)) - torch.log(prop_weights.sum(dim=1))
